__author__ = "Sebastian Maierhofer"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["CAR@TUM"]
__version__ = "2021.2"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Release"

import enum
from typing import List, Optional, Set, Tuple

import numpy as np

import commonroad.geometry.transform
from commonroad.common.validity import *
from commonroad.visualization.drawable import IDrawable
from commonroad.visualization.param_server import ParamServer
from commonroad.visualization.renderer import IRenderer

TRAFFIC_SIGN_VALIDITY_START = {'WARNING_DANGER_SPOT', 'WARNING_RIGHT_BEFORE_LEFT', 'WARNING_STEEP_HILL_DOWNWARDS',
                               'WARNING_SLIPPERY_ROAD', 'WARNING_CONSTRUCTION_SITE', 'WARNING_CROSSING_CYCLIST',
                               'WARNING_ANIMAL_CROSSING_RIGHT', 'RAILWAY', 'PRIORITY_OPPOSITE_DIRECTION',
                               'TURN_RIGHT_AHEAD', 'TURN_LEFT_AHEAD', 'ONEWAY_RIGHT', 'ONEWAY_LEFT',
                               'PRESCRIBED_PASSING_LEFT', 'PRESCRIBED_PASSING_RIGHT', 'BIKEWAY', 'SIDEWALK',
                               'PEDESTRIAN_ZONE_START', 'BICYCLE_ROAD_START', 'BUS_LANE', 'BAN_ALL_VEHICLES',
                               'BAN_CARS', 'BAN_TRUCKS', 'BAN_BICYCLE', 'BAN_MOTORCYCLE', 'BAN_BUS', 'BAN_PEDESTRIAN',
                               'BAN_CAR_TRUCK_BUS_MOTORCYCLE', 'BAN_VEHICLES_CARRYING_DANGEROUS_GOODS', 'NO_ENTRY',
                               'MAX_WEIGHT', 'MAX_WIDTH', 'MAX_HEIGHT', 'MAX_LENGTH', 'MAX_SPEED',
                               'MAX_SPEED_ZONE_START', 'MIN_SPEED',  'NO_OVERTAKING_START',
                               'NO_OVERTAKING_TRUCKS_START', 'TRAFFIC_CALMED_AREA_START', 'PRIORITY_OVER_ONCOMING',
                               'TOWN_SIGN', 'TUNNEL', 'INTERSTATE_START', 'HIGHWAY_START', 'PEDESTRIANS_CROSSING'}

TRAFFIC_SIGN_WITH_ADDITIONAL_VALUE = {'MAX_WEIGHT', 'MAX_WIDTH', 'MAX_HEIGHT', 'MAX_LENGTH', 'MAX_SPEED',
                                      'MAX_SPEED_ZONE_START', 'MIN_SPEED', 'ADDITION_VALID_FOR_X_METERS',
                                      'ADDITION_VALID_IN_X_KILOMETERS', 'ADDITION_TIME_PERIOD_PERMITTED'}

LEFT_HAND_TRAFFIC = {'AUS', 'JPN', 'HKG', 'IND', 'JEY', 'IMN', 'IRL', 'JAM',
                     'KEN', 'MLT', 'MYS', 'NPL', 'NZL', 'ZAF', 'SGP', 'THA',
                     'GBR', 'IDN', 'MAC', 'PAK', 'CYP'}


@enum.unique
class SupportedTrafficSignCountry(enum.Enum):
    GERMANY = 'DEU'
    USA = "USA"
    CHINA = "CHN"
    SPAIN = "ESP"
    RUSSIA = "RUS"
    ARGENTINA = "ARG"
    BELGIUM = "BEL"
    FRANCE = "FRA"
    GREECE = "GRC"
    CROATIA = "HRV"
    ITALY = "ITA"
    PUERTO_RICO = "PRI"
    ZAMUNDA = "ZAM"  # default


@enum.unique
class TrafficSignIDGermany(enum.Enum):
    # default traffic sign IDs (similar to German IDs)
    WARNING_DANGER_SPOT = '101'
    WARNING_RIGHT_BEFORE_LEFT = '102'
    WARNING_LEFT_CURVE = '103-10'
    WARNING_RIGHT_CURVE = '103-20'
    WARNING_STEEP_HILL_DOWNWARDS = '108'
    WARNING_SLIPPERY_ROAD = '114'
    WARNING_CONSTRUCTION_SITE = '123'
    WARNING_TRAFFIC_QUEUES_LIKELY = '124'
    WARNING_ONCOMING_TRAFFIC = '125'
    WARNING_TRAFFIC_LIGHTS_AHEAD = '131'
    WARNING_PEDESTRIANS_RIGHT = '133-10'
    WARNING_PEDESTRIANS_LEFT = '133-20'
    WARNING_CROSSING_CYCLIST = '138'
    WARNING_ANIMAL_CROSSING_RIGHT = '142-10'
    WARNING_LOOSE_GRAVEL = '145-50'
    RAILWAY = '201'
    YIELD = '205'
    STOP = '206'
    PRIORITY_OPPOSITE_DIRECTION = '208'
    TURN_RIGHT_AHEAD = '209-10'
    TURN_LEFT_AHEAD = '209-20'
    KEEP_STRAIGHT_AHEAD = '209-30'
    PRESCRIBED_DIRECTION_RIGHT = '211-20'
    ROUNDABOUT = '215'
    ONEWAY_RIGHT = '220-10'
    ONEWAY_LEFT = '220-20'
    PRESCRIBED_PASSING_LEFT = '222-10'
    PRESCRIBED_PASSING_RIGHT = '222-20'
    DO_NOT_DRIVE_ON_SHOULDER_LANE = '223.2'
    DO_NOT_DRIVE_ON_SHOULDER_LANE_2_LANE = '223.2-50'
    DO_NOT_DRIVE_ON_SHOULDER_LANE_3_LANE = '223.2-51'
    BUS_STOP = '224-50'
    BIKEWAY = '237'
    SIDEWALK = '239'
    PEDESTRIAN_AND_BICYCLE_ROAD = '240'
    PEDESTRIAN_ZONE_START = '242.1'
    PEDESTRIAN_ZONE_END = '242.2'
    BICYCLE_ROAD_START = '244.1'
    BICYCLE_ROAD_END = '244.2'
    BUSLANE = '245'
    BAN_ALL_VEHICLES = '250'
    BAN_CARS = '251'
    BAN_TRUCKS = '253'
    BAN_BICYCLE = '254'
    BAN_MOTORCYCLE = '255'
    BAN_BUS = '257-54'
    BAN_PEDESTRIAN = '259'
    BAN_CAR_TRUCK_BUS_MOTORCYCLE = '260'
    BAN_VEHICLES_CARRYING_DANGEROUS_GOODS = '261'
    MAX_WEIGHT = '262'
    MAX_WIDTH = '264'
    MAX_HEIGHT = '265'
    MAX_LENGTH = '266'
    NO_ENTRY = '267'
    ENVIRONMENTAL_ZONE_START = '270.1'
    ENVIRONMENTAL_ZONE_END = '270.2'
    U_TURN = '272'
    MAX_SPEED = '274'
    MAX_SPEED_ZONE_START = '274.1'
    MAX_SPEED_ZONE_END = '274.2'
    MIN_SPEED = '275'
    NO_OVERTAKING_START = '276'
    NO_OVERTAKING_TRUCKS_START = '277'
    MAX_SPEED_END = '278'
    NO_OVERTAKING_END = '280'
    NO_OVERTAKING_TRUCKS_END = '281'
    ALL_MAX_SPEED_AND_OVERTAKING_END = '282'
    NO_STOP_START_RIGHT = '283-10'
    NO_STOP_CENTER_RIGHT = '283-30'
    RESTRICTED_STOP_CENTER_RIGHT = '286-30'
    RIGHT_OF_WAY = '301'
    PRIORITY = '306'
    PRIORITY_OVER_ONCOMING = '308'
    TOWN_SIGN = '310'
    TOWN_SIGN_BACK = '311'
    PARKING_AREA = '314'
    PARKING_AREA_LEFT = '314-10'
    PARKING_AREA_RIGHT = '314-20'
    PARKING_AREA_RIGHT_LEFT = '314-30'
    TRAFFIC_CALMED_AREA_START = '325.1'
    TRAFFIC_CALMED_AREA_END = '325.2'
    TUNNEL = '327'
    EMERYGECNY_STOP = '328'
    INTERSTATE_START = '330.1'
    INTERSTATE_END = '330.2'
    HIGHWAY_START = '331.1'
    HIGHWAY_END = '331.2'
    HIGHWAY_EXIT_WITH_PLACE_NAME = '332'
    EXIT_ROUTE = '332.1'
    HIGHWAY_EXIT = '333'
    EXIT_BUILT_UP = '333-21'
    EXIT_GENERAL = '333-22'
    PEDESTRIANS_CROSSING = '350'
    WATER_PROTECTION_ZONE = '354'
    TRAFFIC_ASSISTANTS = '356'
    DEAD_END = '357'
    POLICE = '363'
    EMERGENCY_CALL_STATION = '365-51'
    GAS_STATION = '365-52'
    CAMP_AND_CARAVAN_SITE = '365-60'
    ATTRACTION_POINT = '386.1'
    TOURISTIC_ROUTE = '386.2'
    NEARBY_ATTRACTION_POINT = '386.3'
    HIGHWAY_INTERSECTION = '406-50'
    DIRECTION_ARROW_SIGN_MULTI = '418-20'
    DIRECTION_ARROW_SIGN_SINGLE = '419-20'
    DIRECTION_SIGN_CONSOLIDATED = '434-50'
    EXPRESSWAY_ARROW_DIRECTION = '430-20'
    ARROW_SIGN_POST_POINT_OF_INTEREST_LEFT = '432-10'
    STATION = '432-20'
    GUIDE_SIGN_TABLE = '434'
    ADVANCE_DIRECTION = '438'
    DIRECTIONS_SIGN = '439'
    EXPRESSWAY_ENTRANCE_DIRECTIONS = '440'
    INTERSTATE_ANNOUNCEMENT = '448'
    INTERSTATE_ADVANCE_DIRECTION = '449'
    HIGHWAY_EXIT_AHEAD_100_METER = '450-50'
    HIGHWAY_EXIT_AHEAD_200_METER = '450-51'
    HIGHWAY_EXIT_AHEAD_300_METER = '450-52'
    EXPRESSWAY_EXIT_100_METRES = '450-53'
    EXPRESSWAY_EXIT_200_METRES = '450-54'
    EXPRESSWAY_EXIT_300_METRES = '450-55'
    INTERSTATE_DISTANCE = '453'
    DETOUR_SKETCH = '458'
    DETOUR_STRAIGHT = '455.1-30'
    DETOUR_ON_DEMAND_LEFT = '460-10'
    DETOUR_ON_DEMAND_GET_IN_LEFT_LANE = '460-12'
    DETOUR_ON_DEMAND_ANNOUNCEMENT_RIGHT = '460-20'
    DETOUR_ON_DEMAND_RIGHT = '460-21'
    DETOUR_ON_DEMAND_GET_IN_RIGHT_LANE = '460-22'
    DETOUR_ON_DEMAND_STRAIGHTFORWARD = '460-30'
    TRANSITION_3_LEFT_2_TRANSITIONED = '501-15'
    TRANSITION_1_LEFT_1_STRAIGHT = '501-16'
    LANE_BOARD_NO_OPPOSITE_TWO_LANES = '521-30'
    THREE_LANES_NO_ONCOMING_LANES = '521-31'
    FOUR_LANES_NO_ONCOMING_LANES = '521-32'
    FIVE_LANES_NO_ONCOMING_LANES = '521-33'
    LANE_BOARD_3_LANES_NO_OPPOSITE_WITH_SIGNS = '525'
    NARROWING_LANES_1_LANE_FROM_RIGHT = '531-10'
    NARROWING_LANES_1_LANE_FROM_LEFT = '531-20'
    NARROWING_LANES_2_LANES_PLUS_1_LEFT = '531-21'
    FOUR_LANES_NO_ONCOMING_TRAFFIC_TWO_RIGHT_LANES_TURN_RIGHT = '533-22'
    MERGING_LANES_1_LANE_PLUS_1_LANE_RIGHT = '550-20'
    BARRIER = '600-35'
    BARRIER_GATE_100_800 = '600-30'
    BARRIER_GATE_100_1200 = '600-31'
    BARRIER_GATE_100_1600 = '600-32'
    BARRIER_GATE_250_1600 = '600-34'
    BARRIER_GATE = '600-38'
    ROAD_WARNING_POST_SCRAPER_BEACON_RIGHT = '605-10'
    ROAD_WARNING_POST_SCRAPER_BEACON_LEFT = '605-20'
    ROAD_WARNING_POST_ARROW_BEACON_RIGHT = '605-11'
    DIRECTION_SIGN_LEFT_SINGLE = '625-10'
    DIRECTION_SIGN_LEFT_SMALL = '625-11'
    DIRECTION_SIGN_LEFT_MEDIUM = '625-12'
    DIRECTION_SIGN_LEFT_LARGE = '625-13'
    DIRECTION_SIGN_RIGHT_SINGLE = '625-20'
    DIRECTION_SIGN_RIGHT_SMALL = '625-21'
    DIRECTION_SIGN_RIGHT_MEDIUM = '625-22'
    DIRECTION_SIGN_RIGHT_LARGE = '625-23'
    WARNING_PANEL_RIGHT = '626-10'
    WARNING_PANEL_LEFT = '626-20'
    WARNING_PANEL_STRAIGHT_BROAD = '626-30'
    WARNING_PANEL_STRAIGHT_HIGH = '626-31'
    GUIDE_SILL_WITH_GUIDE_BEACON_RIGHT = '628-10'
    GUIDE_RAIL_WITH_GUIDE_BEACON_RIGHT = '629-10'
    GUIDE_PANEL_WITH_GUIDE_BEACON_RIGHT = '629-20'
    GREEN_ARROW = '720'
    ADDITION_LEFT_DIRECTION = '1000'
    ADDITION_LEFT_DIRECTION_1 = '1000-10'
    ADDITION_LEFT_DIRECTION_DANGER_POINT = '1000-11'
    ADDITION_RIGHT_DIRECTION_1 = '1000-20'
    ADDITION_RIGHT_DIRECTION_DANGER_POINT = '1000-21'
    ADDITION_BOTH_DIRECTIONS_HORIZONTAL = '1000-30'
    ADDITION_BOTH_DIRECTIONS_VERTICAL = '1000-31'
    ADDITION_VALID_FOR_X_METERS = '1001-30'
    ADDITION_VALID_FOR_X_KILOMETERS = '1001-31'
    ADDITION_LEFT_TURNING_PRIORITY_WITH_OPPOSITE_RIGHT_YIELD = '1002-10'
    ADDITION_LEFT_TRAFFIC_PRIORITY_WITH_STRAIGHT_RIGHT_YIELD = '1002-11'
    ADDITION_LEFT_TURNING_PRIORITY_WITH_OPPOSITE_YIELD = '1002-12'
    ADDITION_LEFT_TURNING_PRIORITY_WITH_RIGHT_YIELD = '1002-13'
    ADDITION_LEFT_TRAFFIC_PRIORITY_WITH_STRAIGHT_YIELD = '1002-14'
    ADDITION_RIGHT_TURNING_PRIORITY_WITH_OPPOSITE_LEFT_YIELD = '1002-20'
    ADDITION_RIGHT_TRAFFIC_PRIORITY_WITH_STRAIGHT_LEFT_YIELD = '1002-21'
    ADDITION_RIGHT_TURNING_PRIORITY_WITH_OPPOSITE_YIELD = '1002-22'
    ADDITION_RIGHT_TURNING_PRIORITY_WITH_LEFT_YIELD = '1002-23'
    ADDITION_RIGHT_TRAFFIC_PRIORITY_WITH_STRAIGHT_YIELD = '1002-24'
    ADDITION_VALID_IN_X_METERS = '1004-30'
    ADDITION_VALID_IN_X_KILOMETERS = '1004-31'
    ADDITION_VALID_IN_400_METRES = '1004-33'
    ADDITION_VALID_IN_600_METRES = '1004-34'
    ADDITION_VALID_IN_2_KILOMETERS = '1004-35'
    ADDITION_OIL_ON_ROAD = '1006-30'
    ADDITION_SMOKE = '1006-31'
    ADDITION_LOOSE_GRAVEL = '1006-32'
    ADDITION_BUILDING_SITE_EXIT = '1006-33'
    ADDITION_DAMAGED_ROAD = '1006-34'
    ADDITION_DIRTY_ROAD = '1006-35'
    ADDITION_DANGER_OF_COLLISION = '1006-36'
    ADDITION_TOAD_MIGRATION = '1006-37'
    ADDITION_DANGER_OF_CONGESTION = '1006-38'
    ADDITION_RESTRICTED_VIEW_DUE_TO_TREES = '1006-39'
    DANGER_INDICATION_SMOKE = '1007-31'
    ADDITION_CHILDREN_PLAYING_ON_ROAD = '1010-10'
    ADDITION_WINTER_SPORTS_ALLOWED = '1010-11'
    ADDITION_TRAILERS_ALLOWED_TO_PARK_MORE_THAN_14_DAYS = '1010-12'
    ADDITION_CARAVANS_ALLOWED_TO_PARK_MORE_THAN_14_DAYS = '1010-13'
    ADDITION_ROLLING_HIGHWAY = '1010-14'
    ADDITION_LOADING_AREA = '1012-30'
    ADDITION_END = '1012-31'
    ADDITION_GET_OFF_BICYCLES = '1012-32'
    ADDITION_NO_MOPEDS = '1012-33'
    ADDITION_GREEN_WAVE_AT_KM_H = '1012-34'
    ADDITION_STOP_HERE_AT_RED = '1012-35'
    ADDITION_NOISE_CONTROL = '1012-36'
    ADDITION_INFLOW_REGULATION = '1012-37'
    ADDITION_SECONDARY_LANE = '1012-38'
    ADDITION_SCHOOL = '1012-50'
    ADDITION_KINDERGARTEN = '1012-51'
    ADDITION_RETIREMENT_HOME = '1012-52'
    ADDITION_HOSPITAL = '1012-53'
    ADDITION_RESIDENTS_PERMITTED = '1020-30'
    ADDITION_BICYCLES_PERMITTED = '1022-10'
    ADDITION_CARS_PERMITTED = '1024-10'
    ADDITION_AGRICULTURE_PERMITTED = '1026-36'
    ADDITION_FOREST_PERMITTED = '1026-37'
    ADDITION_AGRICULTURE_FOREST_PERMITTED = '1026-38'
    ADDITION_GREEN_STICKER_PERMITTED = '1031-52'
    ADDITION_TIME_PERIOD_PERMITTED = '1040-30'
    ADDITION_MOTOR_VEHICLES_ALLOWED_MASS_3_5_TONS = '1048-12'
    ADDITION_MIN_MASS_3_5_TONS = '1049-13'
    ADDITION_NO_WATER_POLLUTANTS_LOADED = '1052-31'
    ALLOWED_MASS_7_5_TONS = '1053-33'
    ADDITION_VALID_ON_SHOULDER = '1053-34'
    ADDITION_VALID_WHEN_WET = '1053-35'
    UNKNOWN = ''


@enum.unique
class TrafficSignIDZamunda(enum.Enum):  # default traffic sign IDs (similar to German IDs)
    WARNING_DANGER_SPOT = '101'
    WARNING_RIGHT_BEFORE_LEFT = '102'
    WARNING_STEEP_HILL_DOWNWARDS = '108'
    WARNING_SLIPPERY_ROAD = '114'
    WARNING_CONSTRUCTION_SITE = '123'
    WARNING_CROSSING_CYCLIST = '138'
    WARNING_ANIMAL_CROSSING_RIGHT = '142-10'
    RAILWAY = '201'
    YIELD = '205'
    STOP = '206'
    PRIORITY_OPPOSITE_DIRECTION = '208'
    TURN_RIGHT_AHEAD = '209-10'
    TURN_LEFT_AHEAD = '209-20'
    ROUNDABOUT = '215'
    ONEWAY_RIGHT = '220-10'
    ONEWAY_LEFT = '220-20'
    PRESCRIBED_PASSING_LEFT = '222-10'
    PRESCRIBED_PASSING_RIGHT = '222-20'
    BIKEWAY = '237'
    SIDEWALK = '239'
    PEDESTRIAN_ZONE_START = '242.1'
    PEDESTRIAN_ZONE_END = '242.2'
    BICYCLE_ROAD_START = '244.1'
    BICYCLE_ROAD_END = '244.2'
    BUSLANE = '245'
    BAN_ALL_VEHICLES = '250'
    BAN_CARS = '251'
    BAN_TRUCKS = '253'
    BAN_BICYCLE = '254'
    BAN_MOTORCYCLE = '255'
    BAN_BUS = '257-54'
    BAN_PEDESTRIAN = '259'
    BAN_CAR_TRUCK_BUS_MOTORCYCLE = '260'
    BAN_VEHICLES_CARRYING_DANGEROUS_GOODS = '261'
    MAX_WEIGHT = '262'
    MAX_WIDTH = '264'
    MAX_HEIGHT = '265'
    MAX_LENGTH = '266'
    NO_ENTRY = '267'
    U_TURN = '272'
    MAX_SPEED = '274'
    MAX_SPEED_ZONE_START = '274.1'
    MAX_SPEED_ZONE_END = '274.2'
    MIN_SPEED = '275'
    NO_OVERTAKING_START = '276'
    NO_OVERTAKING_TRUCKS_START = '277'
    MAX_SPEED_END = '278'
    NO_OVERTAKING_TRUCKS_END = '281'
    ALL_MAX_SPEED_AND_OVERTAKING_END = '282'
    RIGHT_OF_WAY = '301'
    PRIORITY = '306'
    PRIORITY_OVER_ONCOMING = '308'
    TOWN_SIGN = '310'
    TRAFFIC_CALMED_AREA_START = '325.1'
    TRAFFIC_CALMED_AREA_END = '325.2'
    TUNNEL = '327'
    INTERSTATE_START = '330.1'
    INTERSTATE_END = '330.2'
    HIGHWAY_START = '331.1'
    HIGHWAY_END = '331.2'
    EXIT_BUILT_UP = '333-21'
    EXIT_GENERAL = '333-22'
    PEDESTRIANS_CROSSING = '350'
    DEAD_END = '357'
    DIRECTION_SIGN_LEFT_SINGLE = '625-10'
    DIRECTION_SIGN_LEFT_SMALL = '625-11'
    DIRECTION_SIGN_LEFT_MEDIUM = '625-12'
    DIRECTION_SIGN_LEFT_LARGE = '625-13'
    DIRECTION_SIGN_RIGHT_SINGLE = '625-20'
    DIRECTION_SIGN_RIGHT_SMALL = '625-21'
    DIRECTION_SIGN_RIGHT_MEDIUM = '625-22'
    DIRECTION_SIGN_RIGHT_LARGE = '625-23'
    WARNING_PANEL_RIGHT = '626-10'
    WARNING_PANEL_LEFT = '626-20'
    WARNING_PANEL_STRAIGHT_BROAD = '626-30'
    WARNING_PANEL_STRAIGHT_HIGH = '626-31'
    GREEN_ARROW = '720'
    ADDITION_LEFT_DIRECTION_1 = '1000-10'
    ADDITION_LEFT_DIRECTION_DANGER_POINT = '1000-11'
    ADDITION_RIGHT_DIRECTION_1 = '1000-20'
    ADDITION_RIGHT_DIRECTION_DANGER_POINT = '1000-21'
    ADDITION_BOTH_DIRECTIONS_HORIZONTAL = '1000-30'
    ADDITION_BOTH_DIRECTIONS_VERTICAL = '1000-31'
    ADDITION_VALID_FOR_X_METERS = '1001-30'
    ADDITION_VALID_FOR_X_KILOMETERS = '1001-31'
    ADDITION_LEFT_TURNING_PRIORITY_WITH_OPPOSITE_RIGHT_YIELD = '1002-10'
    ADDITION_LEFT_TRAFFIC_PRIORITY_WITH_STRAIGHT_RIGHT_YIELD = '1002-11'
    ADDITION_LEFT_TURNING_PRIORITY_WITH_OPPOSITE_YIELD = '1002-12'
    ADDITION_LEFT_TURNING_PRIORITY_WITH_RIGHT_YIELD = '1002-13'
    ADDITION_LEFT_TRAFFIC_PRIORITY_WITH_STRAIGHT_YIELD = '1002-14'
    ADDITION_RIGHT_TURNING_PRIORITY_WITH_OPPOSITE_LEFT_YIELD = '1002-20'
    ADDITION_RIGHT_TRAFFIC_PRIORITY_WITH_STRAIGHT_LEFT_YIELD = '1002-21'
    ADDITION_RIGHT_TURNING_PRIORITY_WITH_OPPOSITE_YIELD = '1002-22'
    ADDITION_RIGHT_TURNING_PRIORITY_WITH_LEFT_YIELD = '1002-23'
    ADDITION_RIGHT_TRAFFIC_PRIORITY_WITH_STRAIGHT_YIELD = '1002-24'
    ADDITION_VALID_IN_X_METERS = '1004-30'
    ADDITION_VALID_IN_X_KILOMETERS = '1004-31'
    ADDITION_RESIDENTS_PERMITTED = '1020-30'
    ADDITION_BICYCLES_PERMITTED = '1022-10'
    ADDITION_CARS_PERMITTED = '1024-10'
    ADDITION_AGRICULTURE_PERMITTED = '1026-36'
    ADDITION_FOREST_PERMITTED = '1026-37'
    ADDITION_AGRICULTURE_FOREST_PERMITTED = '1026-38'
    ADDITION_TIME_PERIOD_PERMITTED = '1040-30'
    ALLOWED_MASS_7_5_TONS = '1053-33'
    UNKNOWN = ''


@enum.unique
class TrafficSignIDUsa(enum.Enum):
    MAX_SPEED = 'R2-1'
    U_TURN = 'R3-4'
    ROAD_WORK_AHEAD = 'CW20-1'
    NO_PARKING_ANY_TIME = 'R7-1'
    NO_STANDING = 'R7-4'
    TOW_AWAY_ZONE = 'R7-201a'
    ONE_WAY_LEFT = 'R6-1L'
    ONE_WAY_RIGHT = 'R6-1R'
    DO_NOT_ENTER = 'R5-1'
    NO_LEFT_TURN = 'R3-2'
    RIGHT_TURN_ONLY = 'R3-5R'
    TURN_ONLY_LANES = 'R3-8b'
    NO_RIGHT_TURN = 'R3-1'
    KEEP_RIGHT = 'R4-7'
    SIGNAL_AHEAD = 'W3-3'
    LOADING_ZONE = 'R8-3gP'
    NO_PARKING = 'R8-3'
    LEFT_TURN_ONLY = 'R3-5L'
    NO_STRAIGHT_THROUGH = 'R3-27'
    LEFT_REVERSE_TURN = 'W1-3L'
    PEDESTRIAN_WARNING = 'W11-2'
    DIAGONAL_DOWNWARD_LEFT_ARROW = 'M6-2aL'
    RIGHT_LANE_ENDS = 'W4-2R'
    RESERVED_HANDICAP_PARKING = 'R7-8'
    NO_PARKING_BUS_STOP = 'R7-107'
    ON_PAVEMENT = 'R8-3C'
    DO_NOT_BLOCK_INTERSECTION = 'R10-7'
    WARNING_ARROW_LEFT = 'W1-6L'
    UNKNOWN = ''


@enum.unique
class TrafficSignIDChina(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''


@enum.unique
class TrafficSignIDSpain(enum.Enum):
    YIELD = 'r1'
    STOP = 'r2'
    BAN_ALL_VEHICLES = 'r100'
    NO_ENTRY = 'r101'
    BAN_TRUCKS = 'r106'
    MAX_WEIGHT = 'r107'
    MAX_HEIGHT = 'r205'
    MAX_SPEED = 'r301'
    NO_OVERTAKING_START = 'r305'
    NO_STOPPING = 'r307'
    NO_WAITING = 'r308'
    PEDESTRIANS_CROSSING = 's13'
    UNKNOWN = ''


@enum.unique
class TrafficSignIDRussia(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''  # TODO: add actual IDs


@enum.unique
class TrafficSignIDArgentina(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''  # TODO: add actual IDs


@enum.unique
class TrafficSignIDBelgium(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''  # TODO: add actual IDs


@enum.unique
class TrafficSignIDFrance(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''  # TODO: add actual IDs


@enum.unique
class TrafficSignIDGreece(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''  # TODO: add actual IDs


@enum.unique
class TrafficSignIDCroatia(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''


@enum.unique
class TrafficSignIDItaly(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''


@enum.unique
class TrafficSignIDPuertoRico(enum.Enum):
    MAX_SPEED = '274'  # TODO: change to actual ID
    UNKNOWN = ''  # TODO: add actual IDs


@enum.unique
class TrafficLightDirection(enum.Enum):
    """
    Enum for all the possible directions for a traffic signal
    """
    RIGHT = "right"
    STRAIGHT = "straight"
    LEFT = "left"
    LEFT_STRAIGHT = "leftStraight"
    STRAIGHT_RIGHT = "straightRight"
    LEFT_RIGHT = "leftRight"
    ALL = "all"


@enum.unique
class TrafficLightState(enum.Enum):
    """
    Enum for the possible types of traffic light in signals
    """
    RED = "red"
    YELLOW = "yellow"
    RED_YELLOW = "redYellow"
    GREEN = "green"
    INACTIVE = "inactive"


class TrafficSignElement:
    """ Class which represents a collection of traffic signs at one position"""
    def __init__(self, traffic_sign_element_id: Union[TrafficSignIDZamunda, TrafficSignIDUsa, TrafficSignIDSpain,
                                                      TrafficSignIDGermany, TrafficSignIDChina, TrafficSignIDRussia],
                 additional_values: List[str] = []):
        """

        :param traffic_sign_element_id: ID of traffic sign element (must be element of a traffic sign element enum)
        :param additional_values: list of additional values of a traffic sign element, e.g. velocity, time, city name
        """
        self._traffic_sign_element_id = traffic_sign_element_id
        self._additional_values = additional_values

    @property
    def traffic_sign_element_id(self) -> enum:
        return self._traffic_sign_element_id

    @property
    def additional_values(self) -> List[str]:
        return self._additional_values

    def __eq__(self, other: 'TrafficSignElement'):
        if self.traffic_sign_element_id == other.traffic_sign_element_id \
                and self.additional_values == other.additional_values:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self. _traffic_sign_element_id) + str(self.additional_values))

    def __str__(self):
        return f"Sign Element with id {self._traffic_sign_element_id} and values {self._additional_values} "

    def __repr__(self):
        return f"Sign Element with id {self._traffic_sign_element_id} and values {self._additional_values} "


class TrafficSign(IDrawable):
    """Class to represent a traffic sign"""

    def __init__(self, traffic_sign_id: int,
                 traffic_sign_elements: List[TrafficSignElement],
                 first_occurrence: Set[int], position: np.ndarray,
                 virtual: bool = False):
        """
        :param traffic_sign_id: ID of traffic sign
        :param traffic_sign_elements: list of traffic sign elements
        :param first_occurrence: lanelet ID where traffic sign first appears
        :param position: position of traffic sign
        :param virtual: boolean indicating if this traffic sign is also
        placed there in the real environment or it
        is added for other reasons (e.g., completeness of scenario)
        """
        self._traffic_sign_id = traffic_sign_id
        self._position = position
        self._traffic_sign_elements = traffic_sign_elements
        self._virtual = virtual
        self._first_occurrence = first_occurrence

    @property
    def traffic_sign_id(self) -> int:
        return self._traffic_sign_id

    @property
    def position(self) -> Union[None,np.ndarray]:
        return self._position

    @property
    def traffic_sign_elements(self) -> List[TrafficSignElement]:
        return self._traffic_sign_elements

    @property
    def virtual(self) -> bool:
        return self._virtual

    @property
    def first_occurrence(self) -> Set[int]:
        return self._first_occurrence

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """
        This method translates and rotates a traffic sign

        :param translation: The translation given as [x_off,y_off] for the x and y translation
        :param angle: The rotation angle in radian (counter-clockwise defined)
        """

        assert is_real_number_vector(translation, 2), '<TrafficSign/translate_rotate>: argument translation is ' \
                                                      'not a vector of real ' \
                                                      'numbers of length 2.'
        assert is_real_number(
            angle), '<TrafficSign/translate_rotate>: argument angle must be a ' \
                    'scalar. ' \
                    'angle = %s' % angle
        assert is_valid_orientation(
                angle), '<TrafficSign/translate_rotate>: argument angle must ' \
                        'be ' \
                        'within the ' \
                        'interval [-2pi, 2pi]. angle = %s' % angle
        self._position = commonroad.geometry.transform.translate_rotate(
                np.array([self._position]), translation, angle)[0]

    def __str__(self):
        return f"Sign At {self._position} with {self._traffic_sign_elements} "

    def draw(self, renderer: IRenderer,
             draw_params: Union[ParamServer, dict, None] = None,
             call_stack: Optional[Tuple[str, ...]] = tuple()):
        renderer.draw_traffic_light_sign(self, draw_params, call_stack)


class TrafficLightCycleElement:
    """Class to represent a traffic light cycle"""
    def __init__(self, state: TrafficLightState, duration: int):
        """
        :param state: state of a traffic light cycle element
        :param duration: duration of traffic light cycle element
        """
        self._state = state
        self._duration = duration

    @property
    def state(self) -> TrafficLightState:
        return self._state

    @property
    def duration(self) -> int:
        return self._duration


class TrafficLight(IDrawable):
    """ Class to represent a traffic light"""

    def __init__(self, traffic_light_id: int,
                 cycle: List[TrafficLightCycleElement], position: np.ndarray,
                 time_offset: int = 0,
                 direction: TrafficLightDirection = TrafficLightDirection.ALL,
                 active: bool = True):
        """
        :param traffic_light_id: ID of traffic light
        :param cycle: list of traffic light cycle elements
        :param time_offset: offset of traffic light cycle
        :param position: position of traffic light
        :param direction: driving directions for which the traffic light is valid
        :param active: boolean indicating if traffic light is currently active
        """
        self._traffic_light_id = traffic_light_id
        if len(cycle) == 0:
            self._cycle = get_default_cycle()
        else:
            self._cycle = cycle
        self._time_offset = time_offset
        self._position = position
        self._direction = direction
        self._active = active

    @property
    def traffic_light_id(self) -> int:
        return self._traffic_light_id

    @property
    def cycle(self) -> List[TrafficLightCycleElement]:
        return self._cycle

    def get_state_at_time_step(self, time_step: int) -> TrafficLightState:
        time_step_mod = ((time_step - self.time_offset) % (self.cycle_init_timesteps[-1] - self.time_offset))\
                        + self.time_offset
        i_cycle = np.argmax(time_step_mod < self.cycle_init_timesteps) - 1
        return self.cycle[i_cycle].state

    @property
    def cycle_init_timesteps(self):
        if not hasattr(self, '_cycle_init_timesteps'):
            durations = [cycle_el.duration for cycle_el in self._cycle]
            self._cycle_init_timesteps = np.cumsum(durations) + self.time_offset
            self._cycle_init_timesteps = np.insert(self._cycle_init_timesteps, 0, self.time_offset)
        return self._cycle_init_timesteps

    @property
    def time_offset(self) -> int:
        return self._time_offset

    @property
    def position(self) -> np.ndarray:
        return self._position

    @property
    def direction(self) -> TrafficLightDirection:
        return self._direction

    @direction.setter
    def direction(self, direction: TrafficLightDirection):
        self._direction = direction

    @property
    def active(self) -> bool:
        return self._active

    def translate_rotate(self, translation: np.ndarray, angle: float):
        """
        This method translates and rotates a traffic light

        :param translation: The translation given as [x_off,y_off] for the x and y translation
        :param angle: The rotation angle in radian (counter-clockwise defined)
        """

        assert is_real_number_vector(translation,
                                     2), '<TrafficLight/translate_rotate>: ' \
                                         'argument translation is ' \
                                         'not a vector of real numbers of ' \
                                         'length 2.'
        assert is_real_number(
                angle), '<TrafficLight/translate_rotate>: argument angle must ' \
                        'be ' \
                        'a scalar. ' \
                        'angle = %s' % angle
        assert is_valid_orientation(
                angle), '<TrafficLight/translate_rotate>: argument angle must ' \
                        'be ' \
                        'within the ' \
                        'interval [-2pi, 2pi]. angle = %s' % angle
        self._position = commonroad.geometry.transform.translate_rotate(
                np.array([self._position]), translation, angle)[0]

    def draw(self, renderer: IRenderer,
             draw_params: Union[ParamServer, dict, None] = None,
             call_stack: Optional[Tuple[str, ...]] = tuple()):
        renderer.draw_traffic_light_sign(self, draw_params, call_stack)


def get_default_cycle():
    """
    Defines default traffic light cycle in case no cycle is provided

    _:returns traffic light cycle element
    """
    cycle = [(TrafficLightState.RED, 60),
             (TrafficLightState.RED_YELLOW, 10),
             (TrafficLightState.GREEN, 60),
             (TrafficLightState.YELLOW, 10)]
    cycle_element_list = [TrafficLightCycleElement(state[0], state[1]) for state in cycle]
    return cycle_element_list



import itertools
import math
import os
from collections import defaultdict
from typing import Dict, Callable, Tuple, Union, Any, Set
import commonroad.geometry.shape
import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.patches as patches
import matplotlib.collections as collections
# import pylustrator

import commonroad.prediction.prediction
import commonroad.scenario.obstacle
import commonroad.visualization.draw_dispatch_cr
from commonroad.common.util import Interval
from commonroad.geometry.shape import *
from commonroad.scenario.intersection import Intersection
from commonroad.scenario.traffic_sign import TrafficSign, \
    TrafficLight, \
    TrafficLightState
from commonroad.visualization.traffic_sign import draw_traffic_light_signs
from matplotlib.path import Path
from commonroad.prediction.prediction import Occupancy
from commonroad.scenario.lanelet import LaneletNetwork, Lanelet, LineMarking
from commonroad.scenario.obstacle import DynamicObstacle, \
    StaticObstacle, \
    ObstacleRole, \
    EnvironmentObstacle, \
    PhantomObstacle
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.trajectory import Trajectory, State

from commonroad.visualization.util import draw_polygon_as_patch, \
    draw_polygon_collection_as_patch, \
    LineDataUnits, \
    collect_center_line_colors

__author__ = "Moritz Klischat"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = [""]
__version__ = "2021.2"
__maintainer__ = "Moritz Klischat"
__email__ = "commonroad@lists.lrz.de"
__status__ = "Released"

traffic_sign_path = os.path.join(os.path.dirname(__file__), 'traffic_signs/')


def create_default_draw_params() -> dict:
    basic_shape_parameters_static = {
            'opacity':   1.0,
            'facecolor': '#d95558',
            'edgecolor': '#831d20',
            'linewidth': 0.5,
            'zorder':    20
    }

    basic_shape_parameters_dynamic = {
            'opacity':   1.0,
            'facecolor': '#1d7eea',
            'edgecolor': '#0066cc',
            'linewidth': 0.5,
            'zorder':    20
    }

    draw_params = {
            'scenario': {
                    'dynamic_obstacle':   {
                            'draw_shape':        True,
                            'draw_icon':         False,
                            'draw_bounding_box': True,
                            'show_label':        False,  # show id of obstacle
                            'zorder':            20,
                            'draw_signals':      True,  # draw signal states
                            'signal_radius':     0.5,  # size of a signal states
                            'indicator_color':   '#ebc200',  # turn light
                            'braking_color':     'red',  # braking light
                            'blue_lights_color': 'blue',
                            'horn_color':        'red',
                            # horn is visualized as red center dot
                            'initial_state':     {
                                    'draw_initial_state': False,
                                    # visualize initial state by arrow
                                    # proportional to velocity
                                    'scale_factor':       0.3,
                                    # length of arrow in m per m/s
                                    'kwargs':             {
                                            'linewidth':            1.5,
                                            'length_includes_head': True,
                                            'edgecolor':            'black',
                                            'facecolor':            'black',
                                    },  # further parameters for FancyArrow
                            },
                            'occupancy':         {
                                    'draw_occupancies': 0,
                                    # -1= never, 0= if prediction of vehicle
                                    # is set-based, 1=always
                                    'shape':            {
                                            'polygon':      {
                                                    'opacity':   0.2,
                                                    'facecolor': '#1d7eea',
                                                    'edgecolor': '#0066cc',
                                                    'linewidth': 0.5,
                                                    'zorder':    18,
                                            }, 'rectangle': {
                                                    'opacity':   0.2,
                                                    'facecolor': '#1d7eea',
                                                    'edgecolor': '#0066cc',
                                                    'linewidth': 0.5,
                                                    'zorder':    18,
                                            }, 'circle':    {
                                                    'opacity':   0.2,
                                                    'facecolor': '#1d7eea',
                                                    'edgecolor': '#0066cc',
                                                    'linewidth': 0.5,
                                                    'zorder':    18,
                                            }
                                    },
                            },
                            'shape':             {
                                    'polygon':   basic_shape_parameters_dynamic,
                                    'rectangle': basic_shape_parameters_dynamic,
                                    'circle':    basic_shape_parameters_dynamic
                            },
                            'trajectory':        {
                                    'draw_trajectory': True,
                                    'facecolor':       '#000000',
                                    'draw_continuous': False,
                                    'unique_colors':   False,
                                    'line_width':      0.17,
                                    'z_order':         24
                            }
                    }, 'static_obstacle': {
                            'shape': {
                                    'polygon':   basic_shape_parameters_static,
                                    'rectangle': basic_shape_parameters_static,
                                    'circle':    basic_shape_parameters_static,
                            }
                    }, 'lanelet_network': {
                            'kwargs_traffic_light_signs': {},
                            # further properties for AnnotationBox of traffic
                            # signs or lights, see #
                            # https://matplotlib.org/3.1.0/gallery/text_labels_and_annotations/demo_annotation_box.html
                            'traffic_light':              {
                                    'draw_traffic_lights': True,
                                    'red_color':           'red',
                                    'yellow_color':        '#feb609',
                                    'green_color':         '#00aa16',
                                    'red_yellow_color':    '#fe4009ff',
                                    'show_label':          False,
                                    'scale_factor':        0.25,
                                    'zorder':              30
                            }, 'traffic_sign':            {
                                    'draw_traffic_signs': False,
                                    'show_traffic_signs': 'all',
                                    # 'all' or list of TrafficSignIDs
                                    'speed_limit_unit':   'auto',
                                    # 'mph', 'kmh', 'ms', 'auto'
                                    'show_label':         False,
                                    'scale_factor':       0.25,
                                    'zorder':             30
                            }, 'intersection':            {
                                    'draw_intersections':        False,
                                    'draw_incoming_lanelets':    True,
                                    'incoming_lanelets_color':   '#3ecbcf',
                                    'draw_crossings':            True,
                                    'crossings_color':           '#b62a55',
                                    'draw_successors':           True,
                                    'successors_left_color':     '#ff00ff',
                                    'successors_straight_color': 'blue',
                                    'successors_right_color':    '#ccff00',
                                    'show_label':                False,
                                    # show incoming id and incoming left
                            }, 'lanelet':                 {
                                    'left_bound_color':         '#555555',
                                    'right_bound_color':        '#555555',
                                    'center_bound_color':       '#dddddd',
                                    'unique_colors':            False,
                                    # colorizes center_vertices and labels of
                                    # each lanelet differently
                                    'draw_stop_line':           True,
                                    'stop_line_color':          '#ffffff',
                                    'draw_line_markings':       True,
                                    'draw_left_bound':          True,
                                    'draw_right_bound':         True,
                                    'draw_center_bound':        True,
                                    'draw_border_vertices':     False,
                                    'draw_start_and_direction': True,
                                    'show_label':               False,
                                    # show lanelet_id
                                    'draw_linewidth':           0.5,
                                    'fill_lanelet':             True,
                                    'facecolor':                '#c7c7c7'
                            }
                    },
            }
    }
    # ensure that parameters are also available on higher levels
    draw_params.update(draw_params['scenario'])
    draw_params['shape'] = basic_shape_parameters_static
    draw_params['shape'].update(
            draw_params['scenario']['static_obstacle']['shape'])
    draw_params['occupancy'] = draw_params['scenario']['dynamic_obstacle'][
        'occupancy']
    draw_params['static_obstacle'] = draw_params['scenario']['static_obstacle']
    draw_params['dynamic_obstacle'] = draw_params['scenario'][
        'dynamic_obstacle']
    draw_params['trajectory'] = draw_params['scenario']['dynamic_obstacle'][
        'trajectory']
    draw_params['lanelet_network'] = draw_params['scenario']['lanelet_network']
    draw_params['lanelet'] = draw_params['scenario']['lanelet_network'][
        'lanelet']
    draw_params['intersection'] = draw_params['scenario']['lanelet_network'][
        'intersection']
    draw_params['traffic_sign'] = draw_params['scenario']['lanelet_network'][
        'traffic_sign']
    draw_params['traffic_light'] = draw_params['scenario']['lanelet_network'][
        'traffic_light']
    draw_params['scenario']['lanelet'] = \
        draw_params['scenario']['lanelet_network']['lanelet']

    return draw_params


def line_marking_to_linestyle(line_marking: LineMarking) -> Tuple:
    """:returns: Tuple[line_style, dashes, line_width] for matplotlib
    plotting options."""
    return {
            LineMarking.DASHED:       ('--', (10, 10), 0.25,),
            LineMarking.SOLID:        ('-', (None, None), 0.25),
            LineMarking.BROAD_DASHED: ('--', (10, 10), 0.5),
            LineMarking.BROAD_SOLID:  ('-', (None, None), 0.5)
    }[line_marking]


def traffic_light_color_dict(traffic_light_state: TrafficLightState,
                             params: dict):
    """Retrieve color code for traffic light state."""
    return {
            TrafficLightState.RED:        params['red_color'],
            TrafficLightState.YELLOW:     params['yellow_color'],
            TrafficLightState.GREEN:      params['green_color'],
            TrafficLightState.RED_YELLOW: params['red_yellow_color']
    }[traffic_light_state]


def draw_scenario(obj: Scenario,
                  plot_limits: Union[List[Union[int, float]], None],
                  ax: mpl.axes.Axes, draw_params: dict,
                  draw_func: Dict[type, Callable],
                  handles: Dict[int, List[mpl.patches.Patch]],
                  call_stack: Tuple[str, ...]) -> None:
    """
    :param obj: object to be plotted
    :param plot_limits: draw only objects inside limits [x_ min, x_max,
    y_min, y_max]
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    call_stack = tuple(list(call_stack) + ['scenario'])

    commonroad.visualization.draw_dispatch_cr.draw_object(obj.lanelet_network,
                                                          plot_limits, ax,
                                                          draw_params,
                                                          draw_func, handles,
                                                          call_stack)

    # draw only obstacles inside plot limits
    if plot_limits is not None:
        time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, tuple(), ('time_begin',))
        # dynamic obstacles
        commonroad.visualization.draw_dispatch_cr.draw_object(
                obj.obstacles_by_position_intervals(
                        [Interval(plot_limits[0], plot_limits[1]),
                         Interval(plot_limits[2], plot_limits[3])],
                        tuple([ObstacleRole.DYNAMIC]), time_begin), None, ax,
                draw_params, draw_func, handles, call_stack)
        # static obstacles
        commonroad.visualization.draw_dispatch_cr.draw_object(
                obj.obstacles_by_position_intervals(
                        [Interval(plot_limits[0], plot_limits[1]),
                         Interval(plot_limits[2], plot_limits[3])],
                        tuple([ObstacleRole.STATIC])), None, ax, draw_params,
                draw_func, handles, call_stack)
    else:
        commonroad.visualization.draw_dispatch_cr.draw_object(
                obj.dynamic_obstacles, None, ax, draw_params, draw_func,
                handles, call_stack)
        commonroad.visualization.draw_dispatch_cr.draw_object(
                obj.static_obstacles, None, ax, draw_params, draw_func, handles,
                call_stack)


def draw_static_obstacles(obj: Union[List[StaticObstacle], StaticObstacle],
                          plot_limits: Union[List[Union[int, float]], None],
                          ax: mpl.axes.Axes, draw_params: dict,
                          draw_func: Dict[type, Callable],
                          handles: Dict[int, List[mpl.patches.Patch]],
                          call_stack: Tuple[str, ...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
            draw_params, tuple(), ('time_begin',))

    if type(obj) is StaticObstacle:
        obj = [obj]

    call_stack = tuple(list(call_stack) + ['static_obstacle'])

    shape_list = list()
    for obstacle in obj:
        if type(obstacle) is not StaticObstacle:
            warnings.warn(
                    '<visualization/scenario> Only lists with objects of the '
                    'same '
                    'type can be plotted', UserWarning, stacklevel=1)
            continue
        shape_list.append(obstacle.occupancy_at_time(time_begin).shape)

    collection = shape_batch_dispatcher(shape_list, None, ax, draw_params,
                                        draw_func, handles, call_stack)

    handles.setdefault(StaticObstacle, []).extend(collection)


def draw_environment_obstacles(obj: Union[List[StaticObstacle], StaticObstacle],
                              plot_limits: Union[List[Union[int, float]], None],
                              ax: mpl.axes.Axes, draw_params: dict,
                              draw_func: Dict[type, Callable],
                              handles: Dict[int, List[mpl.patches.Patch]],
                              call_stack: Tuple[str, ...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
        draw_params, tuple(), ('time_begin',))

    if type(obj) is EnvironmentObstacle:
        obj = [obj]

    call_stack = tuple(list(call_stack) + ['static_obstacle'])

    shape_list = list()
    for obstacle in obj:
        if type(obstacle) is not EnvironmentObstacle:
            warnings.warn(
                '<visualization/scenario> Only lists with objects of the '
                'same '
                'type can be plotted', UserWarning, stacklevel=1)
            continue
        shape_list.append(obstacle.occupancy_at_time(time_begin).shape)

    collection = shape_batch_dispatcher(shape_list, None, ax, draw_params,
                                        draw_func, handles, call_stack)

    handles.setdefault(EnvironmentObstacle, []).extend(collection)


def draw_trajectories(obj: Union[List[Trajectory], Trajectory],
                      plot_limits: Union[List[Union[int, float]], None],
                      ax: mpl.axes.Axes, draw_params: dict,
                      draw_func: Dict[type, Callable],
                      handles: Dict[int, List[mpl.patches.Patch]],
                      call_stack: Tuple[str, ...]) -> List[
    mpl.collections.Collection]:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """

    if type(obj) is Trajectory:
        obj = [obj]

    try:
        facecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('trajectory', 'facecolor'))
        unique_colors = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('trajectory', 'unique_colors'))
        line_width = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('trajectory', 'line_width'))
        draw_continuous = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('trajectory', 'draw_continuous'))
        z_order = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('trajectory', 'z_order'))
        time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_begin',))
        time_end = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['time_end']))
    except KeyError:
        print("Cannot find stylesheet for trajectory. Called through:")
        print(call_stack)

    if time_begin == time_end:
        return []

    # select unique colors from colormap for each lanelet's center_line
    colormap = None
    if unique_colors is True:
        norm = mpl.colors.Normalize(vmin=0, vmax=len(obj))
        colormap = cm.ScalarMappable(norm=norm, cmap=cm.brg)

    patchlist = list()
    traj_list = list()

    for traj in obj:
        traj_points = list()
        for time_step in range(time_begin, time_end):
            tmp = traj.state_at_time_step(time_step)
            if tmp is not None:
                traj_points.append(tmp.position)
            else:
                if time_begin > traj.initial_time_step:
                    break

        if traj_points:
            traj_list.append(traj_points)

    if len(traj_list) > 0:
        if draw_continuous is True:
            paths = []
            for traj in traj_list:
                paths.append(Path(traj, closed=False))
            if unique_colors is True:
                collection = []
                for i_p, path in enumerate(paths):
                    facecolor = colormap.to_rgba(i_p)
                    collection.append(
                            collections.PathCollection([path], color=facecolor,
                                                       lw=line_width,
                                                       zorder=z_order,
                                                       facecolor='none'))
                    ax.add_collection(collection[-1])
            else:
                collection = [collections.PathCollection(paths, color=facecolor,
                                                         lw=line_width,
                                                         zorder=z_order,
                                                         facecolor='none')]
        else:
            if unique_colors is True:
                collection = []
                for i_p, traj in enumerate(traj_list):
                    facecolor = colormap.to_rgba(i_p)
                    traj = np.array(traj)
                    collection.append(collections.EllipseCollection(
                            np.ones([traj.shape[0], 1]) * line_width,
                            np.ones([traj.shape[0], 1]) * line_width,
                            np.zeros([traj.shape[0], 1]), offsets=traj,
                            units='xy', linewidths=0, zorder=z_order,
                            transOffset=ax.transData, facecolor=facecolor))
                    ax.add_collection(collection[-1])
            else:
                try:
                    traj_list = np.array(np.concatenate(traj_list))
                except:
                    dsdf = 0
                collection = [collections.EllipseCollection(
                        np.ones([traj_list.shape[0], 1]) * line_width,
                        np.ones([traj_list.shape[0], 1]) * line_width,
                        np.zeros([traj_list.shape[0], 1]), offsets=traj_list,
                        units='xy', linewidths=0, zorder=z_order,
                        transOffset=ax.transData, facecolor=facecolor)]
        if draw_continuous is False:
            ax.add_collection(collection[0])
    else:
        collection = []

    return collection


def draw_lanelet_network(obj: LaneletNetwork,
                         plot_limits: Union[List[Union[int, float]], None],
                         ax: mpl.axes.Axes, draw_params: dict,
                         draw_func: Dict[type, Callable],
                         handles: Dict[int, List[mpl.patches.Patch]],
                         call_stack: Tuple[str, ...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    call_stack = tuple(list(call_stack) + ['lanelet_network'])
    # if plot_limits is not None:
    #     center = [plot_limits[1]+plot_limits[0]*0.5, plot_limits[
    #     3]+plot_limits[2]*0.5]
    #     radius = np.sqrt((plot_limits[1]-plot_limits[0])**2 + (plot_limits[
    #     3]-plot_limits[2])**2)
    #     lanelets = obj.lanelets_in_proximity(center, radius)
    # else:

    _draw_lanelets_intersection(obj, obj._traffic_lights, obj._traffic_signs,
                                obj.intersections, None, ax, draw_params,
                                draw_func, handles, call_stack)


def draw_lanelet_list(obj: List[Lanelet],
                      plot_limits: Union[List[Union[int, float]], None],
                      ax: mpl.axes.Axes, draw_params: dict,
                      draw_func: Dict[type, Callable],
                      handles: Dict[int, List[mpl.patches.Patch]],
                      call_stack: Tuple[str, ...]) -> None:
    """
    Draws list of lanelets.
    """
    if isinstance(obj, Lanelet):
        obj = [obj]

    _draw_lanelets_intersection(LaneletNetwork.create_from_lanelet_list(obj),
                                None, None, None, plot_limits, ax, draw_params,
                                draw_func, handles, call_stack)


def _draw_lanelets_intersection(obj: LaneletNetwork, traffic_lights: Union[
    Dict[int, TrafficLight], None], traffic_signs: Union[
    Dict[int, TrafficSign], None],
                                intersections: Union[List[Intersection], None],
                                plot_limits: Union[
                                    List[Union[int, float]], None],
                                ax: mpl.axes.Axes, draw_params: dict,
                                draw_func: Dict[type, Callable],
                                handles: Dict[int, List[mpl.patches.Patch]],
                                call_stack: Tuple[str, ...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    lanelets = obj.lanelets

    try:
        time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_begin',))
        if traffic_lights is not None:
            draw_traffic_lights = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'traffic_light', 'draw_traffic_lights'))

            traffic_light_colors = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'traffic_light'))
        else:
            draw_traffic_lights = False

        if traffic_signs is not None:
            draw_traffic_signs = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'traffic_sign', 'draw_traffic_signs'))
            show_traffic_sign_label = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'traffic_sign', 'show_label'))
        else:
            draw_traffic_signs = show_traffic_sign_label = False

        if intersections is not None and len(intersections) > 0:
            draw_intersections = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'intersection', 'draw_intersections'))
        else:
            draw_intersections = False

        if draw_intersections is True:
            draw_incoming_lanelets = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack, ('lanelet_network', 'intersection',
                                              'draw_incoming_lanelets'))
            incoming_lanelets_color = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack, ('lanelet_network', 'intersection',
                                              'incoming_lanelets_color'))
            draw_crossings = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'intersection', 'draw_crossings'))
            crossings_color = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'intersection', 'crossings_color'))
            draw_successors = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'intersection', 'draw_successors'))
            successors_left_color = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack, ('lanelet_network', 'intersection',
                                              'successors_left_color'))
            successors_straight_color = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack, ('lanelet_network', 'intersection',
                                              'successors_straight_color'))
            successors_right_color = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack, ('lanelet_network', 'intersection',
                                              'successors_right_color'))
            show_intersection_labels = \
                commonroad.visualization.draw_dispatch_cr._retrieve_value(
                    draw_params, call_stack,
                    ('lanelet_network', 'intersection', 'show_label'))
        else:
            draw_incoming_lanelets = draw_crossings = draw_successors = \
                show_intersection_labels = False

        left_bound_color = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'left_bound_color'))
        right_bound_color = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('lanelet', 'right_bound_color'))
        center_bound_color = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('lanelet', 'center_bound_color'))
        unique_colors = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'unique_colors'))
        draw_stop_line = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'draw_stop_line'))
        stop_line_color = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'stop_line_color'))
        draw_line_markings = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('lanelet', 'draw_line_markings'))
        show_label = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('lanelet', 'show_label'))
        draw_border_vertices = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('lanelet', 'draw_border_vertices'))
        draw_left_bound = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'draw_left_bound'))
        draw_right_bound = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'draw_right_bound'))
        draw_center_bound = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('lanelet', 'draw_center_bound'))
        draw_start_and_direction = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack,
                ('lanelet', 'draw_start_and_direction'))
        draw_linewidth = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'draw_linewidth'))
        fill_lanelet = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('lanelet', 'fill_lanelet'))
        facecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('lanelet', 'facecolor'))
        antialiased = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, tuple(), tuple(['antialiased']))

    except KeyError:
        print("Cannot find stylesheet for lanelet. Called through:")
        print(call_stack)

    incoming_lanelets = set()
    incomings_left = {}
    incomings_id = {}
    crossings = set()
    all_successors = set()
    successors_left = set()
    successors_straight = set()
    successors_right = set()
    if draw_intersections:
        # collect incoming lanelets
        if draw_incoming_lanelets:
            incomings: List[set] = [incoming.incoming_lanelets for intersection
                                    in intersections for incoming in
                                    intersection.incomings]
            incoming_lanelets: Set[int] = set.union(*incomings)
            for intersection in intersections:
                for incoming in intersection.incomings:
                    for l_id in incoming.incoming_lanelets:
                        incomings_left[l_id] = incoming.left_of
                        incomings_id[l_id] = incoming.incoming_id

        if draw_crossings:
            tmp_list: List[set] = [intersection.crossings for intersection in
                                   intersections]
            crossings: Set[int] = set.union(*tmp_list)

        if draw_successors:
            tmp_list: List[set] = [incoming.successors_left for intersection in
                                   intersections for incoming in
                                   intersection.incomings]
            successors_left: Set[int] = set.union(*tmp_list)
            tmp_list: List[set] = [incoming.successors_straight for intersection
                                   in intersections for incoming in
                                   intersection.incomings]
            successors_straight: Set[int] = set.union(*tmp_list)
            tmp_list: List[set] = [incoming.successors_right for intersection in
                                   intersections for incoming in
                                   intersection.incomings]
            successors_right: Set[int] = set.union(*tmp_list)
            all_successors = set.union(successors_straight, successors_right,
                                       successors_left)

    # direction arrow
    codes_direction = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]

    scale_direction = 1.5
    pts = np.array([[0.0, -0.5, 1.0], [1.0, 0.0, 1.0], [0.0, 0.5, 1.0],
                    [0.0, -0.5, 1.0]])
    scale_m = np.array(
            [[scale_direction, 0, 0], [0, scale_direction, 0], [0, 0, 1]])

    def direction(x, y, angle):
        """Returns path of arrow shape"""
        transform = np.array([[np.cos(angle), -np.sin(angle), x],
                              [np.sin(angle), np.cos(angle), y], [0, 0, 1]])
        ptr_trans = transform.dot(scale_m.dot(pts.transpose()))
        ptr_trans = ptr_trans[0:2, :]
        ptr_trans = ptr_trans.transpose()

        path = Path(ptr_trans, codes_direction)
        return path

    # select unique colors from colormap for each lanelet's center_line
    colormap = None
    if unique_colors is True:
        norm = mpl.colors.Normalize(vmin=0, vmax=len(lanelets))
        colormap = cm.ScalarMappable(norm=norm, cmap=cm.jet)

    incoming_vertices_fill = list()
    crossing_vertices_fill = list()
    succ_left_paths = list()
    succ_straight_paths = list()
    succ_right_paths = list()

    vertices_fill = list()
    coordinates_left_border_vertices = np.empty((0, 2))
    coordinates_right_border_vertices = np.empty((0, 2))
    direction_list = list()
    center_paths = list()
    left_paths = list()
    right_paths = list()

    if draw_traffic_lights:
        center_line_color_dict = collect_center_line_colors(obj, traffic_lights,
                                                            time_begin)

    # collect paths for drawing
    for i_lanelet, lanelet in enumerate(lanelets):
        if unique_colors:
            # set center bound color to unique value
            center_bound_color = colormap.to_rgba(i_lanelet)

        if draw_start_and_direction:
            center = lanelet.center_vertices[0]
            tan_vec = np.array(lanelet.right_vertices[0]) - np.array(
                    lanelet.left_vertices[0])
            if unique_colors is False:
                direction_list.append(direction(center[0], center[1],
                                                np.arctan2(tan_vec[1], tan_vec[
                                                    0]) + 0.5 * np.pi))
            else:
                direction_list = [direction(center[0], center[1],
                                            np.arctan2(tan_vec[1], tan_vec[
                                                0]) + 0.5 * np.pi)]
                ax.add_collection(collections.PathCollection(direction_list,
                                                             color=center_bound_color,
                                                             lw=0.5,
                                                             zorder=10.1,
                                                             antialiased=antialiased))

        # left bound
        if draw_border_vertices or draw_left_bound:
            if draw_border_vertices:
                coordinates_left_border_vertices = np.vstack((
                        coordinates_left_border_vertices,
                        lanelet.left_vertices))

            if draw_line_markings and lanelet.line_marking_left_vertices is \
                    not LineMarking.UNKNOWN and \
                    lanelet.line_marking_left_vertices is not \
                    LineMarking.NO_MARKING:
                linestyle, dashes, linewidth_metres = line_marking_to_linestyle(
                        lanelet.line_marking_left_vertices)
                tmp_left = lanelet.left_vertices.copy()
                tmp_left[0, :] = \
                    lanelet.interpolate_position(linewidth_metres / 2)[2]
                tmp_left[-1, :] = lanelet.interpolate_position(
                        lanelet.distance[-1] - linewidth_metres / 2)[2]
                line = LineDataUnits(tmp_left[:, 0], tmp_left[:, 1], zorder=12,
                                     linewidth=linewidth_metres, alpha=1.0,
                                     color=left_bound_color,
                                     linestyle=linestyle, dashes=dashes)
                ax.add_line(line)
            else:
                left_paths.append(Path(lanelet.left_vertices, closed=False))

        # right bound
        if draw_border_vertices or draw_right_bound:

            if draw_border_vertices:
                coordinates_right_border_vertices = np.vstack((
                        coordinates_right_border_vertices,
                        lanelet.right_vertices))

            if draw_line_markings and lanelet.line_marking_right_vertices is \
                    not LineMarking.UNKNOWN and \
                    lanelet.line_marking_right_vertices is not \
                    LineMarking.NO_MARKING:
                linestyle, dashes, linewidth_metres = line_marking_to_linestyle(
                        lanelet.line_marking_right_vertices)
                tmp_right = lanelet.right_vertices.copy()
                tmp_right[0, :] = \
                    lanelet.interpolate_position(linewidth_metres / 2)[1]
                tmp_right[-1, :] = lanelet.interpolate_position(
                        lanelet.distance[-1] - linewidth_metres / 2)[1]
                line = LineDataUnits(tmp_right[:, 0], tmp_right[:, 1],
                                     zorder=10.5, linewidth=linewidth_metres,
                                     alpha=1.0, color=right_bound_color,
                                     linestyle=linestyle, dashes=dashes)
                ax.add_line(line)
            else:
                right_paths.append(Path(lanelet.right_vertices, closed=False))

        # stop line
        if draw_stop_line and lanelet.stop_line:
            stop_line = np.vstack(
                    [lanelet.stop_line.start, lanelet.stop_line.end])
            linestyle, dashes, linewidth_metres = line_marking_to_linestyle(
                    lanelet.stop_line.line_marking)
            # cut off in the beginning, because linewidth_metres is added later
            vec = stop_line[1, :] - stop_line[0, :]
            tangent = vec / np.linalg.norm(vec)
            stop_line[0, :] += linewidth_metres * tangent / 2
            stop_line[1, :] -= linewidth_metres * tangent / 2
            line = LineDataUnits(stop_line[:, 0], stop_line[:, 1], zorder=11,
                                 linewidth=linewidth_metres, alpha=1.0,
                                 color=stop_line_color, linestyle=linestyle,
                                 dashes=dashes)
            ax.add_line(line)

        # visualize traffic light state through colored center bound
        has_traffic_light = draw_traffic_lights and lanelet.lanelet_id in \
                            center_line_color_dict
        if has_traffic_light:
            light_state = center_line_color_dict[lanelet.lanelet_id]

            if light_state is not TrafficLightState.INACTIVE:
                linewidth_metres = 0.75
                # dashed line for red_yellow
                linestyle = '--' if light_state == \
                                    TrafficLightState.RED_YELLOW else '-'
                dashes = (5, 5) if linestyle == '--' else (None, None)

                # cut off in the beginning, because linewidth_metres is added
                # later
                tmp_center = lanelet.center_vertices.copy()
                if lanelet.distance[-1] > linewidth_metres:
                    tmp_center[0, :] = \
                        lanelet.interpolate_position(linewidth_metres)[0]
                zorder = 10.05 if light_state == TrafficLightState.GREEN else\
                    10.0
                line = LineDataUnits(tmp_center[:, 0], tmp_center[:, 1],
                                     zorder=zorder, linewidth=linewidth_metres,
                                     alpha=0.7,
                                     color=traffic_light_color_dict(light_state,
                                                                    traffic_light_colors),
                                     linestyle=linestyle, dashes=dashes)
                ax.add_line(line)

        # draw colored center bound. Hierarchy or colors: successors > usual
        # center bound
        is_successor = draw_intersections and draw_successors and \
                       lanelet.lanelet_id in all_successors
        if is_successor:
            if lanelet.lanelet_id in successors_left:
                succ_left_paths.append(
                        Path(lanelet.center_vertices, closed=False))
            elif lanelet.lanelet_id in successors_straight:
                succ_straight_paths.append(
                        Path(lanelet.center_vertices, closed=False))
            else:
                succ_right_paths.append(
                        Path(lanelet.center_vertices, closed=False))

        elif draw_center_bound:
            if unique_colors is False:
                center_paths.append(Path(lanelet.center_vertices, closed=False))
            else:
                center_paths = [Path(lanelet.center_vertices, closed=False)]
                ax.add_collection(collections.PathCollection(center_paths,
                                                             edgecolor=center_bound_color,
                                                             facecolor='none',
                                                             lw=draw_linewidth,
                                                             zorder=10,
                                                             antialiased=antialiased))

        is_incoming_lanelet = draw_intersections and draw_incoming_lanelets \
                              and (
                lanelet.lanelet_id in incoming_lanelets)
        is_crossing = draw_intersections and draw_crossings and (
                lanelet.lanelet_id in crossings)
        if fill_lanelet:
            if not is_incoming_lanelet and not is_crossing:
                vertices_fill.append(np.concatenate((lanelet.right_vertices,
                                                     np.flip(
                                                             lanelet.left_vertices,
                                                             0))))

        # collect incoming lanelets in separate list for plotting in
        # different color
        if is_incoming_lanelet:
            incoming_vertices_fill.append(np.concatenate((
                    lanelet.right_vertices, np.flip(lanelet.left_vertices, 0))))
        elif is_crossing:
            crossing_vertices_fill.append(np.concatenate((
                    lanelet.right_vertices, np.flip(lanelet.left_vertices, 0))))

        # DRAW LABELS INTO LANELET CENTER
        if show_label or show_intersection_labels or draw_traffic_signs:
            strings = []
            if show_label:
                strings.append(str(lanelet.lanelet_id))
            if is_incoming_lanelet and show_intersection_labels:
                strings.append(
                        'inc_id: ' + str(incomings_id[lanelet.lanelet_id]))
                strings.append(
                        'inc_left: ' + str(incomings_left[lanelet.lanelet_id]))
            if draw_traffic_signs and show_traffic_sign_label is True:
                traffic_signs_tmp = [traffic_signs[id] for id in
                                     lanelet.traffic_signs]
                if traffic_signs_tmp:
                    # add as text to label
                    str_tmp = 'traffic signs: '
                    add_str = ''
                    for sign in traffic_signs_tmp:
                        for el in sign.traffic_sign_elements:
                            # TrafficSignIDGermany(
                            # el.traffic_sign_element_id).name would give the
                            # name
                            str_tmp += add_str + \
                                       el.traffic_sign_element_id.value
                            add_str = ', '

                    strings.append(str_tmp)

            label_string = ', '.join(strings)
            if len(label_string) > 0:
                # compute normal angle of label box
                clr_positions = lanelet.interpolate_position(
                        0.5 * lanelet.distance[-1])
                normal_vector = np.array(clr_positions[1]) - np.array(
                        clr_positions[2])
                angle = np.rad2deg(
                        np.arctan2(normal_vector[1], normal_vector[0])) - 90
                angle = angle if Interval(-90, 90).contains(
                        angle) else angle - 180

                ax.text(clr_positions[0][0], clr_positions[0][1], label_string,
                        bbox={'facecolor': center_bound_color, 'pad': 2},
                        horizontalalignment='center',
                        verticalalignment='center', rotation=angle, zorder=30.2)

    # draw paths and collect axis handles
    if draw_right_bound:
        ax.add_collection(collections.PathCollection(right_paths,
                                                     edgecolor=right_bound_color,
                                                     facecolor='none',
                                                     lw=draw_linewidth,
                                                     zorder=10,
                                                     antialiased=antialiased))
    if draw_left_bound:
        ax.add_collection(collections.PathCollection(left_paths,
                                                     edgecolor=left_bound_color,
                                                     facecolor='none',
                                                     lw=draw_linewidth,
                                                     zorder=10,
                                                     antialiased=antialiased))
    if unique_colors is False:
        if draw_center_bound:
            ax.add_collection(collections.PathCollection(center_paths,
                                                         edgecolor=center_bound_color,
                                                         facecolor='none',
                                                         lw=draw_linewidth,
                                                         zorder=10,
                                                         antialiased=antialiased))
        if draw_start_and_direction:
            ax.add_collection(collections.PathCollection(direction_list,
                                                         color=center_bound_color,
                                                         lw=0.5, zorder=10.1,
                                                         antialiased=antialiased))

    if successors_left:
        ax.add_collection(collections.PathCollection(succ_left_paths,
                                                     edgecolor=successors_left_color,
                                                     facecolor='none',
                                                     lw=draw_linewidth * 3.0,
                                                     zorder=11,
                                                     antialiased=antialiased))
    if successors_straight:
        ax.add_collection(collections.PathCollection(succ_straight_paths,
                                                     edgecolor=successors_straight_color,
                                                     facecolor='none',
                                                     lw=draw_linewidth * 3.0,
                                                     zorder=11,
                                                     antialiased=antialiased))
    if successors_right:
        ax.add_collection(collections.PathCollection(succ_right_paths,
                                                     edgecolor=successors_right_color,
                                                     facecolor='none',
                                                     lw=draw_linewidth * 3.0,
                                                     zorder=11,
                                                     antialiased=antialiased))

    # fill lanelets with facecolor
    collection_tmp = collections.PolyCollection(vertices_fill,
                                                transOffset=ax.transData,
                                                zorder=9.0, facecolor=facecolor,
                                                edgecolor='none',
                                                antialiased=antialiased)
    ax.add_collection(collection_tmp)
    if incoming_vertices_fill:
        collection_tmp = collections.PolyCollection(incoming_vertices_fill,
                                                    transOffset=ax.transData,
                                                    facecolor=incoming_lanelets_color,
                                                    edgecolor='none',
                                                    zorder=9.1,
                                                    antialiased=antialiased)
        ax.add_collection(collection_tmp)
    if crossing_vertices_fill:
        collection_tmp = collections.PolyCollection(crossing_vertices_fill,
                                                    transOffset=ax.transData,
                                                    facecolor=crossings_color,
                                                    edgecolor='none',
                                                    zorder=9.2,
                                                    antialiased=antialiased)
        ax.add_collection(collection_tmp)

    # draw_border_vertices
    if draw_border_vertices:
        # left vertices
        collection_tmp = collections.EllipseCollection(
                np.ones([coordinates_left_border_vertices.shape[0], 1]) * 1,
                np.ones([coordinates_left_border_vertices.shape[0], 1]) * 1,
                np.zeros([coordinates_left_border_vertices.shape[0], 1]),
                offsets=coordinates_left_border_vertices,
                color=left_bound_color, transOffset=ax.transData)
        ax.add_collection(collection_tmp)

        # right_vertices
        collection_tmp = collections.EllipseCollection(
                np.ones([coordinates_right_border_vertices.shape[0], 1]) * 1,
                np.ones([coordinates_right_border_vertices.shape[0], 1]) * 1,
                np.zeros([coordinates_right_border_vertices.shape[0], 1]),
                offsets=coordinates_right_border_vertices,
                color=right_bound_color, transOffset=ax.transData)
        ax.add_collection(collection_tmp)

    traffic_lights_signs = []
    if draw_traffic_signs:
        # draw actual traffic sign
        traffic_lights_signs.extend(list(traffic_signs.values()))

    if draw_traffic_lights:
        # draw actual traffic sign
        traffic_lights_signs.extend(list(traffic_lights.values()))

    if traffic_lights_signs:
        draw_traffic_light_signs(traffic_lights_signs, None, ax, draw_params,
                                 draw_func, handles, call_stack)


def draw_dynamic_obstacles(obj: Union[List[DynamicObstacle], DynamicObstacle],
                           plot_limits: Union[List[Union[int, float]], None],
                           ax: mpl.axes.Axes, draw_params: dict,
                           draw_func: Dict[type, Callable], handles: Dict[
            Any, List[Union[mpl.patches.Patch, mpl.collections.Collection]]],
                           call_stack: Tuple[str, ...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """

    def collecting(o: DynamicObstacle):
        occupancy_list = list()
        trajectory = None
        shape = None
        indicators = []
        braking = []
        horns = []
        bluelights = []

        if type(o) is not DynamicObstacle:
            warnings.warn(
                    '<visualization/scenario> Only lists with objects of the '
                    'same '
                    'type can be plotted', UserWarning, stacklevel=1)
            return (occupancy_list, trajectory, shape)

        # draw occupancies
        if (draw_occupancies == 1 or (draw_occupancies == 0 and type(
                o.prediction) == commonroad.prediction.prediction.SetBasedPrediction)):
            if draw_shape:
                # occupancy already plotted
                time_begin_occ = time_begin + 1
            else:
                time_begin_occ = time_begin

            for time_step in range(time_begin_occ, time_end):
                tmp = o.occupancy_at_time(time_step)
                if tmp is not None:
                    occupancy_list.append(tmp)

        # get trajectory
        if draw_trajectory and type(
                o.prediction) == \
                commonroad.prediction.prediction.TrajectoryPrediction:
            trajectory = o.prediction.trajectory

        # get shape
        if draw_shape:
            occ = o.occupancy_at_time(time_begin)
            if occ is None:
                shape = None
            else:
                shape = occ.shape

            # draw signal states
            if draw_signals and occ is not None:
                sig = o.signal_state_at_time_step(time_begin)
                if sig is not None:
                    if isinstance(occ.shape, Rectangle):
                        if hasattr(sig,
                                   'hazard_warning_lights') and \
                                sig.hazard_warning_lights is True:
                            indicators.extend([occ.shape.vertices[0],
                                               occ.shape.vertices[1],
                                               occ.shape.vertices[2],
                                               occ.shape.vertices[3]])
                        else:
                            if hasattr(sig,
                                       'indicator_left') and \
                                    sig.indicator_left is True:
                                indicators.extend([occ.shape.vertices[1],
                                                   occ.shape.vertices[2]])
                            if hasattr(sig,
                                       'indicator_right') and \
                                    sig.indicator_right is True:
                                indicators.extend([occ.shape.vertices[0],
                                                   occ.shape.vertices[3]])
                        if hasattr(sig,
                                   'braking_lights') and sig.braking_lights \
                                is True:
                            braking.extend([occ.shape.vertices[0],
                                            occ.shape.vertices[1]])
                        if hasattr(sig,
                                   'flashing_blue_lights') and \
                                sig.flashing_blue_lights is True:
                            bluelights.append(occ.shape.center)
                        if hasattr(sig, 'horn') and sig.horn is True:
                            horns.append(occ.shape.center)
                    else:
                        warnings.warn(
                                'Plotting signal states only implemented for '
                                'obstacle_shapes Rectangle.')

        # draw car icon
        if draw_icon and type(
                o.prediction) == \
                commonroad.prediction.prediction.TrajectoryPrediction:
            if time_begin == 0:
                inital_state = o.initial_state
            else:
                inital_state = o.prediction.trajectory.state_at_time_step(
                        time_begin)
            if inital_state is not None:
                draw_car(inital_state.position[0], inital_state.position[1],
                         inital_state.orientation, 2.5, ax, zorder=30)

        initial_state = None
        if show_label:
            if time_begin == 0:
                initial_state = o.initial_state
                initial_position = initial_state.position
                handles.setdefault(DynamicObstacle, []).append(
                        ax.text(initial_position[0] + 0.5, initial_position[1],
                                str(o.obstacle_id), clip_on=True, zorder=1000))
            elif type(
                    o.prediction) == \
                    commonroad.prediction.prediction.TrajectoryPrediction:
                initial_state = o.prediction.trajectory.state_at_time_step(
                        time_begin)
                if initial_state is not None:
                    initial_position = \
                        o.prediction.trajectory.state_at_time_step(
                            time_begin).position
                    handles.setdefault(DynamicObstacle, []).append(
                            ax.text(initial_position[0] + 0.5,
                                    initial_position[1], str(o.obstacle_id),
                                    clip_on=True, zorder=1000))

        if draw_initial_state:
            if initial_state is None:
                if time_begin == 0:
                    initial_state = o.initial_state
                elif type(
                        o.prediction) == \
                        commonroad.prediction.prediction.TrajectoryPrediction:
                    initial_state = o.prediction.trajectory.state_at_time_step(
                            time_begin)

        return [occupancy_list, trajectory, shape,
                np.array(indicators).reshape(-1, 2),
                np.array(braking).reshape(-1, 2),
                np.array(horns).reshape(-1, 2),
                np.array(bluelights).reshape(-1, 2), initial_state]

    if type(obj) is DynamicObstacle:
        obj = [obj]

    try:
        time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_begin',))
        time_end = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_end',))
        draw_icon = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'draw_icon'))
        show_label = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'show_label'))
        draw_shape = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'draw_shape'))
        draw_initial_state = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'initial_state', 'draw_initial_state'))
        scale_factor = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'initial_state', 'scale_factor'))
        kwargs_init_state = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'initial_state', 'kwargs'))
        draw_occupancies = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'occupancy', 'draw_occupancies'))
        draw_signals = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'draw_signals'))
        zorder = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'zorder'))
        signal_radius = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'signal_radius'))
        indicator_color = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'indicator_color'))
        braking_color = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'braking_color'))
        horn_color = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'horn_color'))
        blue_lights_color = \
            commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'blue_lights_color'))
        draw_trajectory = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'trajectory', 'draw_trajectory'))
    except KeyError:
        warnings.warn(
                "Cannot find stylesheet for dynamic_obstacle. Called through:")
        print(call_stack)

    call_stack = tuple(list(call_stack) + ['dynamic_obstacle'])

    # collect objects from all vehicles to draw them efficiently in batches
    occupancy_list, trajectories_list, shapes_list, indicators, braking, \
    horns, bluelights, initial_states = zip(
            *list(map(collecting, obj)))

    occupancy_list = list(filter(None, list(occupancy_list)))
    trajectories_list = list(filter(None, list(trajectories_list)))
    shapes_list = list(filter(None, list(shapes_list)))
    indicators = np.vstack(indicators)
    braking = np.vstack(braking)
    horns = np.vstack(horns)
    bluelights = np.vstack(bluelights)
    initial_states = list(filter(None, list(initial_states)))

    # draw collected lists, store handles:
    if len(shapes_list) > 0:
        handles.setdefault(DynamicObstacle, []).extend(
                shape_batch_dispatcher(shapes_list, None, ax, draw_params,
                                       draw_func, handles, call_stack))
    if len(trajectories_list) > 0:
        handles.setdefault(DynamicObstacle, []).extend(
                commonroad.visualization.draw_dispatch_cr.draw_object(
                        trajectories_list, None, ax, draw_params, draw_func,
                        handles, call_stack))
    if len(occupancy_list) > 0:
        handles.setdefault(DynamicObstacle, []).extend(
                commonroad.visualization.draw_dispatch_cr.draw_object(
                        occupancy_list, None, ax, draw_params, draw_func,
                        handles, call_stack))

    # draw initial state
    if draw_initial_state is True:
        for initial_state in initial_states:
            handles.setdefault(DynamicObstacle, []).extend(
                    draw_state(initial_state, call_stack=call_stack, ax=ax,
                               draw_params=draw_params,
                               scale_factor=scale_factor, plot_limits=None,
                               arrow_args=kwargs_init_state))

    # draw signals
    if indicators.size > 0:
        diameters = signal_radius * np.ones(indicators.shape[0]) * 2
        handles.setdefault(DynamicObstacle, []).append(
                collections.EllipseCollection(diameters, diameters,
                                              angles=np.zeros_like(diameters),
                                              offsets=indicators,
                                              transOffset=ax.transData,
                                              units='xy',
                                              facecolor=indicator_color,
                                              edgecolor=indicator_color,
                                              zorder=zorder + 0.2, linewidth=0))
        ax.add_collection(handles[DynamicObstacle][-1])
    if braking.size > 0:
        diameters = signal_radius * np.ones(braking.shape[0]) * 3.0
        handles.setdefault(DynamicObstacle, []).append(
                collections.EllipseCollection(diameters, diameters,
                                              angles=np.zeros_like(diameters),
                                              offsets=braking,
                                              transOffset=ax.transData,
                                              units='xy',
                                              facecolor=braking_color,
                                              edgecolor=braking_color,
                                              zorder=zorder + 0.1, linewidth=0))
        ax.add_collection(handles[DynamicObstacle][-1])
    if horns.size > 0:
        diameters = signal_radius * np.ones(horns.shape[0]) * 3.0
        handles.setdefault(DynamicObstacle, []).append(
                collections.EllipseCollection(diameters, diameters,
                                              angles=np.zeros_like(diameters),
                                              offsets=horns,
                                              transOffset=ax.transData,
                                              units='xy', facecolor=horn_color,
                                              edgecolor=braking_color,
                                              zorder=zorder + 0.1, linewidth=0))
        ax.add_collection(handles[DynamicObstacle][-1])
    if bluelights.size > 0:
        diameters = signal_radius * np.ones(bluelights.shape[0]) * 2
        handles.setdefault(DynamicObstacle, []).append(
                collections.EllipseCollection(diameters, diameters,
                                              angles=np.zeros_like(diameters),
                                              offsets=bluelights,
                                              transOffset=ax.transData,
                                              units='xy',
                                              facecolor=blue_lights_color,
                                              edgecolor=braking_color,
                                              zorder=zorder + 0.1, linewidth=0))
        ax.add_collection(handles[DynamicObstacle][-1])


def draw_phantom_obstacles(obj: Union[List[PhantomObstacle], PhantomObstacle],
                           plot_limits: Union[List[Union[int, float]], None],
                           ax: mpl.axes.Axes, draw_params: dict,
                           draw_func: Dict[type, Callable], handles: Dict[
            Any, List[Union[mpl.patches.Patch, mpl.collections.Collection]]],
                           call_stack: Tuple[str, ...]) -> None:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """

    def collecting(o: PhantomObstacle):
        occupancy_list = list()

        if type(o) is not PhantomObstacle:
            warnings.warn(
                    '<visualization/scenario> Only lists with objects of the '
                    'same '
                    'type can be plotted', UserWarning, stacklevel=1)
            return (occupancy_list)

        # draw occupancies
        if (draw_occupancies == 1 or (draw_occupancies == 0 and type(
                o.prediction) == commonroad.prediction.prediction.SetBasedPrediction)):
            if draw_shape:
                # occupancy already plotted
                time_begin_occ = time_begin + 1
            else:
                time_begin_occ = time_begin

            for time_step in range(time_begin_occ, time_end):
                tmp = o.occupancy_at_time(time_step)
                if tmp is not None:
                    occupancy_list.append(tmp)

        return [occupancy_list]

    if type(obj) is PhantomObstacle:
        obj = [obj]

    try:
        time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_begin',))
        time_end = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_end',))
        draw_shape = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('dynamic_obstacle', 'draw_shape'))
        draw_occupancies = commonroad.visualization.draw_dispatch_cr\
            ._retrieve_value(
                draw_params, call_stack,
                ('dynamic_obstacle', 'occupancy', 'draw_occupancies'))

    except KeyError:
        warnings.warn(
                "Cannot find stylesheet for dynamic_obstacle (phantom_obstacle. Called through:")
        print(call_stack)

    call_stack = tuple(list(call_stack) + ['phantom_obstacle'])

    # collect objects from all vehicles to draw them efficiently in batches
    occupancy_list = zip(*list(map(collecting, obj)))

    occupancy_list = list(filter(None, list(occupancy_list)))

    # draw collected lists, store handles:
    if len(occupancy_list) > 0:
        handles.setdefault(DynamicObstacle, []).extend(
                commonroad.visualization.draw_dispatch_cr.draw_object(
                        occupancy_list, None, ax, draw_params, draw_func,
                        handles, call_stack))

        ax.add_collection(handles[DynamicObstacle][-1])


def draw_occupancies(obj: Union[List[Occupancy], Occupancy],
                     plot_limits: Union[List[Union[int, float]], None],
                     ax: mpl.axes.Axes, draw_params: dict,
                     draw_func: Dict[type, Callable],
                     handles: Dict[int, List[mpl.patches.Patch]],
                     call_stack: Tuple[str, ...]) -> List[
    mpl.collections.Collection]:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    call_stack = tuple(list(call_stack) + ['occupancy'])

    if type(obj) is Occupancy:
        obj = [obj]

    # collect occupancy shape(s)
    shape_list = list()
    if not 'dynamic_obstacle' in call_stack:
        time_begin = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_begin',))
        time_end = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, ('time_end',))

        for occupancy in obj:
            if time_begin <= occupancy.time_step <= time_end:
                shape_list.append(occupancy.shape)
    else:
        for occupancy in obj:
            shape_list.append(occupancy.shape)

    patch_list = shape_batch_dispatcher(shape_list, None, ax, draw_params,
                                        draw_func, handles, call_stack)

    return patch_list


def draw_shape_group(obj: ShapeGroup,
                     plot_limits: Union[List[Union[int, float]], None],
                     ax: mpl.axes.Axes, draw_params: dict,
                     draw_func: Dict[type, Callable],
                     handles: Dict[int, List[mpl.patches.Patch]],
                     call_stack: Tuple[str, ...]) -> List[mpl.patches.Patch]:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    patch_list: List[mpl.patches.Patch] = list()

    for shape in obj.shapes:
        patch = commonroad.visualization.draw_dispatch_cr.draw_object(shape,
                                                                      None, ax,
                                                                      draw_params,
                                                                      draw_func,
                                                                      handles,
                                                                      call_stack)
        patch_list.extend(patch)

    return patch_list


def shape_batch_dispatcher(obj: List[Shape],
                           plot_limits: Union[List[Union[int, float]], None],
                           ax: mpl.axes.Axes, draw_params: dict,
                           draw_func: Dict[type, Callable],
                           handles: Dict[int, List[mpl.patches.Patch]],
                           call_stack: Tuple[str, ...]) -> List[
    mpl.collections.Collection]:
    """
    Groups a list of shapes by their type and draws them each in a batch.
    :param obj: list of shapes to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    shapes_ordered_by_type = dict()

    # sorts shapes by type
    for shape in obj:
        if shape is not None:
            shapes_ordered_by_type.setdefault(type(shape), []).append(shape)

    collection_list = list()
    # plots each type
    for shape_list_tmp in shapes_ordered_by_type.values():
        collection_tmp = commonroad.visualization.draw_dispatch_cr.draw_object(
                shape_list_tmp, None, ax, draw_params, draw_func, handles,
                call_stack)
        collection_list.extend(collection_tmp)

    return collection_list


def draw_polygons(obj: Polygon,
                  plot_limits: Union[List[Union[int, float]], None],
                  ax: mpl.axes.Axes, draw_params: dict,
                  draw_func: Dict[type, Callable],
                  handles: Dict[int, List[mpl.patches.Patch]],
                  call_stack: Tuple[str, ...]) -> List[
    mpl.collections.Collection]:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    if type(obj) is Polygon:
        obj = [obj]

    call_stack = tuple(list(call_stack) + ['shape', 'polygon'])
    try:
        facecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['facecolor']))
        edgecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['edgecolor']))
        zorder = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['zorder']))
        opacity = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['opacity']))
        linewidth = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['linewidth']))
        antialiased = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, tuple(), tuple(['antialiased']))
    except KeyError:
        print("Cannot find stylesheet for polygon. Called through:")
        print(call_stack)

    vertices_list = list()

    for poly in obj:
        if type(poly) is not Polygon:
            warnings.warn(
                    '<visualization/scenario> Only lists with objects of the '
                    'same '
                    'type can be plotted', UserWarning, stacklevel=1)
            continue
        vertices_list.append(np.array(poly.vertices))
    collection = draw_polygon_collection_as_patch(vertices_list, ax,
                                                  zorder=zorder,
                                                  facecolor=facecolor,
                                                  edgecolor=edgecolor,
                                                  lw=linewidth, alpha=opacity,
                                                  antialiased=antialiased)
    return [collection]


def draw_circle(obj: Union[Circle, List[Circle]],
                plot_limits: Union[List[Union[int, float]], None],
                ax: mpl.axes.Axes, draw_params: dict,
                draw_func: Dict[type, Callable],
                handles: Dict[int, List[mpl.patches.Patch]],
                call_stack: Tuple[str, ...]) -> List[
    mpl.collections.Collection]:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    if type(obj) is Circle:
        obj: List[Circle] = [obj]

    call_stack = tuple(list(call_stack) + ['shape', 'circle'])
    try:
        facecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['facecolor']))
        edgecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['edgecolor']))
        zorder = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['zorder']))
        opacity = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['opacity']))
        linewidth = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['linewidth']))
    except KeyError:
        print("Cannot find stylesheet for circle. Called through:")
        print(call_stack)

    centers = list()
    radii = list()
    for circle in obj:
        if type(circle) is not Circle:
            warnings.warn(
                    '<visualization/scenario> Only lists with objects of the '
                    'same '
                    'type can be plotted', UserWarning, stacklevel=1)
            continue
        centers.append(circle.center)
        radii.append(circle.radius)

    centers = np.array(centers)
    diameters = np.array(radii) * 2
    collection = collections.EllipseCollection(diameters, diameters,
                                               angles=np.zeros_like(radii),
                                               offsets=centers,
                                               transOffset=ax.transData,
                                               units='xy', facecolor=facecolor,
                                               edgecolor=edgecolor,
                                               zorder=zorder,
                                               linewidth=linewidth,
                                               alpha=opacity, )

    ax.add_collection(collection)

    return [collection]


def draw_rectangle(obj: Union[Rectangle, List[Rectangle]],
                   plot_limits: Union[List[Union[int, float]], None],
                   ax: mpl.axes.Axes, draw_params: dict,
                   draw_func: Dict[type, Callable],
                   handles: Dict[int, List[mpl.patches.Patch]],
                   call_stack: Tuple[str, ...]) -> List[
    mpl.collections.Collection]:
    """
    :param obj: object to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    if type(obj) is Rectangle:
        obj = [obj]

    call_stack = tuple(list(call_stack) + ['shape', 'rectangle'])
    try:
        facecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['facecolor']))
        edgecolor = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['edgecolor']))
        zorder = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['zorder']))
        opacity = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['opacity']))
        linewidth = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, call_stack, tuple(['linewidth']))
        antialiased = commonroad.visualization.draw_dispatch_cr._retrieve_value(
                draw_params, tuple(), tuple(['antialiased']))
    except KeyError:
        print("Cannot find stylesheet for rectangle. Called through:")
        print(call_stack)

    vertices = list()

    for rect in obj:
        if type(rect) is not Rectangle:
            warnings.warn(
                    '<visualization/scenario> Only lists with objects of the '
                    'same '
                    'type can be plotted', UserWarning, stacklevel=1)
            continue
        vertices.append(np.array(rect.vertices))

    collection = collections.PolyCollection(vertices, closed=True,
                                            zorder=zorder,
                                            transOffset=ax.transData,
                                            facecolor=facecolor,
                                            edgecolor=edgecolor, alpha=opacity,
                                            antialiased=antialiased,
                                            linewidth=linewidth)
    ax.add_collection(collection)

    return [collection]


def draw_car(pos_x: Union[int, float], pos_y: Union[int, float],
             rotate: Union[int, float], scale: Union[int, float],
             ax: mpl.axes.Axes, zorder: int = 5, carcolor: str = '#ffffff',
             lw=0.5):
    rotate = rotate + np.pi

    def reshape_and_addup(verts):
        verts = verts.reshape((int(len(verts) / 2), 2))
        for i in range(1, len(verts)):
            verts[i] = verts[i] + verts[i - 1]
        return verts

    def transform(vertices, pos_x, pos_y, rotate, scale):
        vertices = np.asarray(np.bmat([vertices, np.ones((len(vertices), 1))]))
        tmat = np.array([[np.cos(rotate), -np.sin(rotate), pos_x],
                         [np.sin(rotate), np.cos(rotate), pos_y], [0, 0, 1]])
        scalemat = np.array(
                [[scale * 1 / 356.3211, 0, 0], [0, scale * 1 / 356.3211, 0],
                 [0, 0, 1]])
        centermat = np.array([[1, 0, -250], [0, 1, -889], [0, 0, 1]])
        tmat = tmat.dot(scalemat.dot(centermat))
        vertices = tmat.dot(vertices.transpose())
        vertices = np.array([[1, 0, 0], [0, 1, 0]]).dot(vertices).transpose()
        return vertices

    verts1 = np.array(
            [193.99383, 752.94359, -22.66602, 38, -9.33398, 52.66797, -2.60743,
             44.82812, -0.0586, 0, 0.0293, 0.91992, -0.0293, 0.91797, 0.0586, 0,
             2.60743, 44.82813, 9.33398, 52.66797, 22.66602, 38.00003, 59.33398,
             -7.334, 62, -10.666, -6, -50.00003, -2.65234, -68.41407, 2.65234,
             -68.41601, 6, -50, -62, -10.66602])

    verts2 = np.array(
            [715.99381, 768.27757, -101.332, 6.66602, 10.666, 62.66797, 3.3223,
             50.41797, -3.3223, 50.41796, -10.666, 62.66601, 101.332, 6.668, 22,
             -42.66799, 9.9824, -76.83594,  # 0.018,0,
             # -0.01,-0.24804,
             # 0.01,-0.24805,
             # -0.018,0,
             -9.9824, -76.83789, 0, 0])

    verts3 = np.array(
            [421.06111, 751.61113, 190.2667, 3.33333, 13.3333, 5.33334,
             -108.6666, 12.66666, -134, 0, -119.3334, -18.66666, 127.6456,
             -2.96473])

    verts4 = np.array(
            [271.32781, 712.14446, -6, -0.8, -16, 12, -14.8, 19.2, -4, 6, 20.4,
             0.4, 3.6, -4.4, 4.8, -2.8])
    verts5 = np.array(
            [191.32781, 753.94359, -99.999996, 11, -63, 18.5, -59, 38.5, -20,
             77, 20, 59.52734, 57, 36.49998, 65, 20.49999, 99.999996, 11.0001])

    verts6 = np.array([421.06111, 1027.399, 190.2667, -3.3333, 13.3333, -5.3333,
                       -108.6666, -12.6667, -134, 0, -119.3334, 18.6667,
                       127.6456, 2.9647])

    verts7 = np.array(
            [271.32781, 1066.8657, -6, 0.8, -16, -12, -14.8, -19.2, -4, -6,
             20.4, -0.4, 3.6, 4.4, 4.8, 2.8])

    verts8 = np.array(
            [389.79851, 728.34788, -343.652396, 10.16016, -68.666, 22.42969,
             -29.2558, 74.57031, -7.3164, 60.35742, -0.074, 0, 0.037, 0.76758,
             -0.037, 0.76758, 0.074, 0, 7.3164, 42.35937, 29.2558, 74.57031,
             68.666, 22.4278, 343.652396, 10.1621, 259.5859, -4.6192, 130.2539,
             -17.5527, 24.0196, -18.4766, 17.5527, -65.58788, 3.6953, -37.42773,
             0, -13.24414, -3.6953, -55.42774, -17.5527, -65.58984, -24.0196,
             -18.47656, -130.2539, -17.55274])

    verts1 = reshape_and_addup(verts1)  # vorderscheibe
    verts2 = reshape_and_addup(verts2)  # rueckscheibe
    verts3 = reshape_and_addup(verts3)  # seitenscheibe rechts
    verts4 = reshape_and_addup(verts4)  # rueckspiegel links
    verts5 = reshape_and_addup(verts5)  # motorhaube
    verts6 = reshape_and_addup(verts6)  # fenster rechts
    verts7 = reshape_and_addup(verts7)  # rueckspiegel rechts
    verts8 = reshape_and_addup(verts8)  # umriss

    verts1 = transform(verts1, pos_x, pos_y, rotate, scale)
    verts2 = transform(verts2, pos_x, pos_y, rotate, scale)
    verts3 = transform(verts3, pos_x, pos_y, rotate, scale)
    verts4 = transform(verts4, pos_x, pos_y, rotate, scale)
    verts5 = transform(verts5, pos_x, pos_y, rotate, scale)
    verts6 = transform(verts6, pos_x, pos_y, rotate, scale)
    verts7 = transform(verts7, pos_x, pos_y, rotate, scale)
    verts8 = transform(verts8, pos_x, pos_y, rotate, scale)

    windowcolor = '#555555'
    draw_polygon_as_patch(verts1, ax, facecolor=windowcolor, zorder=zorder + 1,
                          lw=lw)
    draw_polygon_as_patch(verts2, ax, facecolor=windowcolor, zorder=zorder + 1,
                          lw=lw)
    draw_polygon_as_patch(verts3, ax, facecolor=windowcolor, zorder=zorder + 1,
                          lw=lw)
    draw_polygon_as_patch(verts4, ax, facecolor='#ffffff', zorder=zorder + 1,
                          lw=lw)
    ax.plot(verts5[:, 0], verts5[:, 1], zorder=zorder + 1, color='#000000',
            lw=lw)
    draw_polygon_as_patch(verts6, ax, facecolor=windowcolor, zorder=zorder + 1,
                          lw=lw)
    draw_polygon_as_patch(verts7, ax, facecolor='#ffffff', zorder=zorder + 1,
                          lw=lw)
    draw_polygon_as_patch(verts8, ax, facecolor=carcolor, edgecolor='#000000',
                          zorder=zorder, lw=lw)


def draw_state(state: State, plot_limits: Union[List[Union[int, float]], None],
               ax: mpl.axes.Axes, draw_params: dict,
               draw_func: Dict[type, Callable] = None,
               handles: Dict[int, List[mpl.patches.Patch]] = None,
               call_stack: Tuple[str, ...] = None, scale_factor=None,
               arrow_args=None) -> List[mpl.collections.Collection]:
    """
    :param state: state to be plotted
    :param ax: axes object from matplotlib
    :param draw_params: parameters for plotting given by a nested dict that
    recreates the structure of an object,
    :param draw_func: specifies the drawing function
    :param call_stack: tuple of string containing the call stack,
    which allows for differentiation of plotting styles
           depending on the call stack of draw_object
    :return: None
    """
    try:
        if scale_factor is None:
            scale_factor = commonroad.visualization.draw_dispatch_cr\
                ._retrieve_value(
                    draw_params, call_stack, ('initial_state', 'scale_factor'))
        if arrow_args is None:
            arrow_args = commonroad.visualization.draw_dispatch_cr\
                ._retrieve_value(
                    draw_params, call_stack, ('initial_state', 'kwargs'))
    except KeyError:
        print("Cannot find stylesheet for initial_state. Called through:")
        print(call_stack)
        return []

    cos = math.cos(state.orientation)
    sin = math.sin(state.orientation)
    length = 1.5
    x = state.position[0] + cos * length
    y = state.position[1] + sin * length
    artist = ax.arrow(x=x, y=y, dx=state.velocity * cos * scale_factor,
                      dy=state.velocity * sin * scale_factor, zorder=100,
                      **arrow_args)
    return [artist]


# default function dict, which assigns drawing functions to object classes
draw_func_dict = {
        commonroad.scenario.scenario.Scenario:             draw_scenario,
        commonroad.scenario.lanelet.Lanelet:               draw_lanelet_list,
        commonroad.scenario.lanelet.LaneletNetwork:        draw_lanelet_network,
        commonroad.scenario.traffic_sign.TrafficSign:      draw_traffic_light_signs,
        commonroad.scenario.traffic_sign.TrafficLight:     draw_traffic_light_signs,
        commonroad.scenario.obstacle.DynamicObstacle:      draw_dynamic_obstacles,
        commonroad.scenario.obstacle.StaticObstacle:       draw_static_obstacles,
        commonroad.scenario.obstacle.PhantomObstacle :     draw_phantom_obstacles,
        commonroad.scenario.obstacle.EnvironmentObstacle:  draw_environment_obstacles,
        commonroad.scenario.trajectory.Trajectory:         draw_trajectories,
        commonroad.geometry.shape.ShapeGroup:              draw_shape_group,
        commonroad.geometry.shape.Polygon:                 draw_polygons,
        commonroad.geometry.shape.Circle:                  draw_circle,
        commonroad.geometry.shape.Rectangle:               draw_rectangle,
        commonroad.prediction.prediction.Occupancy:        draw_occupancies
}
